"""
Utilities for IceProd functionality.
"""

from __future__ import absolute_import, division, print_function


class NoncriticalError(Exception):
    """An exception that can be logged and then ignored."""
    def __init__(self, value=''):
        self.value = value

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.value:
            return 'NoncriticalError(%r)'%(self.value)
        else:
            return 'NoncriticalError()'

    def __reduce__(self):
        return (NoncriticalError,(self.value,))



"""
Suspend any jobs and tasks from datasets that are suspended / errors.
"""

import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('dataset_cleanup')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    datasets = await rest_client.request('GET', '/dataset_summaries/status')
    dataset_ids = []
    if 'suspended' in datasets:
        dataset_ids.extend(list(datasets['suspended']))
    if 'errors' in datasets:
        dataset_ids.extend(list(datasets['errors']))
    for dataset_id in dataset_ids:
        try:
            jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/job_summaries/status')
            if 'processing' in jobs:
                for job_id in jobs['processing']:
                    args = {'status': 'suspended'}
                    await rest_client.request('PUT', f'/datasets/{dataset_id}/jobs/{job_id}/status', args)

            tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/task_summaries/status')
            task_ids = []
            for status in tasks:
                if status not in ('complete','suspended','failed'):
                    task_ids.extend(tasks[status])
            while task_ids:
                tids = task_ids[:10000]
                task_ids = task_ids[10000:]
                args = {'tasks': tids}
                await rest_client.request('POST', f'/datasets/{dataset_id}/task_actions/bulk_status/suspended', args)

        except Exception:
            logger.error('error cleaning a job/task in dataset %s', dataset_id, exc_info=True)
            if debug:
                raise

    # clean up tasks in suspended/errors jobs in processing datasets
    dataset_ids = []
    if 'processing' in datasets:
        dataset_ids.extend(list(datasets['processing']))
    for dataset_id in dataset_ids:
        try:
            jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/job_summaries/status')
            job_ids = set()
            if 'suspended' in jobs:
                job_ids.update(jobs['suspended'])
            if 'errors' in jobs:
                job_ids.update(jobs['errors'])

            args = {
                'keys': 'task_id|job_id|status'
            }
            tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks', args)
            task_ids = []
            for tasks in tasks.values():
                if tasks['status'] not in ('complete','suspended') and tasks['job_id'] in job_ids:
                    task_ids.append(tasks['task_id'])
            while task_ids:
                tids = task_ids[:10000]
                task_ids = task_ids[10000:]
                args = {'tasks': tids}
                await rest_client.request('POST', f'/datasets/{dataset_id}/task_actions/bulk_status/suspended', args)

        except Exception:
            logger.error('error cleaning a task in dataset %s', dataset_id, exc_info=True)
            if debug:
                raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Update pilots when tasks change state via a user.
"""

import argparse
import asyncio
import logging

import requests.exceptions

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('removed_tasks')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    try:
        async def test_pilot(pilot):
            new_tasks = []
            for task_id in pilot['tasks']:
                ret = await rest_client.request('GET', f'/tasks/{task_id}')
                if ret['status'] == 'processing':
                    new_tasks.append(task_id)
            if new_tasks != pilot['tasks']:
                args = {'tasks': new_tasks}
                try:
                    await rest_client.request('PATCH', f'/pilots/{pilot["pilot_id"]}', args)
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code != 404:
                        raise

        awaitables = set()
        pilots = await rest_client.request('GET', '/pilots?keys=pilot_id|tasks')
        for p in pilots.values():
            if 'tasks' in p and p['tasks']:
                awaitables.add(asyncio.create_task(test_pilot(p)))
                if len(awaitables) >= 40:
                    done,pending = await asyncio.wait(awaitables, return_when=asyncio.FIRST_COMPLETED)
                    awaitables = pending
                    for fut in done:
                        await fut
        for fut in asyncio.as_completed(awaitables):
            await fut

    except Exception:
        logger.error('error updating pilot for removed tasks', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)
    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""Functions to communicate with the server using JSONRPC"""

from __future__ import absolute_import, division, print_function

import sys
import os
import time
import json
from copy import deepcopy
from datetime import datetime
import gzip
import logging

from rest_tools.client import RestClient

from iceprod.core import constants
from iceprod.core import functions
from iceprod.core import dataclasses
from iceprod.core.resources import Resources
from .serialization import dict_to_dataclasses


class ServerComms:
    """
    Setup JSONRPC communications with the IceProd server.

    Args:
        url (str): address to connect to
        passkey (str): passkey for authorization/authentication
        config (:py:class:`iceprod.server.exe.Config`): Config object
        **kwargs: passed to JSONRPC
    """
    def __init__(self, url, passkey, config, **kwargs):
        self.rest = RestClient(address=url,token=passkey,**kwargs)

    async def download_task(self, gridspec, resources={}, site='', query_params=None):
        """
        Download new task(s) from the server.

        Args:
            gridspec (str): gridspec the pilot was submitted from
            resources (dict): resources available in the pilot
            site (str): site where pilot is running

        Returns:
            list: list of task configs
        """
        hostname = functions.gethostname()
        domain = '.'.join(hostname.split('.')[-2:])
        try:
            ifaces = functions.getInterfaces()
        except Exception:
            ifaces = None
        resources = deepcopy(resources)
        if 'gpu' in resources and isinstance(resources['gpu'],list):
            resources['gpu'] = len(resources['gpu'])
        os_type = os.environ['OS_ARCH'] if 'OS_ARCH' in os.environ else None
        if os_type:
            resources['os'] = os_type
        if site:
            resources['site'] = site
        args = {
            'gridspec': gridspec,
            'hostname': hostname,
            'domain': domain,
            'ifaces': ifaces,
            'requirements': resources,
        }
        if query_params:
            args['query_params'] = query_params
        task = await self.rest.request('POST', '/task_actions/process', args)
        if not task:
            return None

        # get config
        try:
            config = await self.rest.request('GET', '/config/{}'.format(task['dataset_id']))
            if not isinstance(config, dataclasses.Job):
                config = dict_to_dataclasses(config)
        except Exception:
            logging.warning('failed to get dataset config for dataset %s', task['dataset_id'])
            await self.task_kill(task['task_id'], dataset_id=task['dataset_id'],
                                 reason='failed to download dataset config')
            raise

        # fill in options
        if 'options' not in config:
            config['options'] = {}
        config['options']['task_id'] = task['task_id']
        config['options']['job_id'] = task['job_id']
        config['options']['dataset_id'] = task['dataset_id']
        config['options']['task'] = task['task_index']
        if 'requirements' in task:
            config['options']['resources'] = {k:task['requirements'][k] for k in Resources.defaults}
        try:
            job = await self.rest.request('GET', '/jobs/{}'.format(task['job_id']))
            config['options']['job'] = job['job_index']
        except Exception:
            logging.warning('failed to get job %s', task['job_id'])
            await self.task_kill(task['task_id'], dataset_id=task['dataset_id'],
                                 reason='failed to download job')
            raise
        try:
            dataset = await self.rest.request('GET', '/datasets/{}'.format(task['dataset_id']))
            config['options']['dataset'] = dataset['dataset']
            config['options']['jobs_submitted'] = dataset['jobs_submitted']
            config['options']['tasks_submitted'] = dataset['tasks_submitted']
            config['options']['debug'] = dataset['debug']
        except Exception:
            logging.warning('failed to get dataset %s', task['dataset_id'])
            await self.task_kill(task['task_id'], dataset_id=task['dataset_id'],
                                 reason='failed to download dataset')
            raise
        return [config]

    async def task_files(self, dataset_id, task_id):
        """
        Get the task files for a dataset and task.

        Args:
            dataset_id (str): dataset_id
            task_id (str): task_id

        Returns:
            list: list of :py:class:`iceprod.core.dataclasses.Data` objects
        """
        ret = await self.rest.request('GET', '/datasets/{}/files/{}'.format(dataset_id, task_id))
        data = []
        for r in ret['files']:
            d = dataclasses.Data(r)
            if not d.valid():
                raise Exception('returned Data not valid')
            data.append(d)
        return data

    async def processing(self, task_id):
        """
        Tell the server that we are processing this task.

        Only used for single task config, not for pilots.

        Args:
            task_id (str): task_id to mark as processing
        """
        await self.rest.request('PUT', '/tasks/{}/status'.format(task_id), {'status': 'processing'})

    async def finish_task(self, task_id, dataset_id=None, stats={},
                          stat_filter=None, start_time=None, resources=None,
                          site=None):
        """
        Finish a task.

        Args:
            task_id (str): task_id of task
            dataset_id (str): (optional) dataset_id of task
            stats (dict): (optional) task statistics
            stat_filter (iterable): (optional) stat filter by keywords
            start_time (float): (optional) task start time in unix seconds
            resources (dict): (optional) task resource usage
            site (str): (optional) site the task is running at
        """
        if stat_filter:
            # filter task stats
            stats = {k:stats[k] for k in stats if k in stat_filter}

        hostname = functions.gethostname()
        domain = '.'.join(hostname.split('.')[-2:])
        if start_time:
            t = time.time() - start_time
        elif resources and 'time' in resources and resources['time']:
            t = int(resources['time']*3600)
        else:
            t = None
        iceprod_stats = {
            'hostname': hostname,
            'domain': domain,
            'time_used': t,
            'task_stats': stats,
            'time': datetime.utcnow().isoformat(),
        }
        if resources:
            iceprod_stats['resources'] = resources
        if site:
            iceprod_stats['site'] = site
        if dataset_id:
            iceprod_stats['dataset_id'] = dataset_id

        await self.rest.request('POST', '/tasks/{}/task_stats'.format(task_id), iceprod_stats)

        data = {}
        if t:
            data['time_used'] = t
        if site:
            data['site'] = site
        await self.rest.request('POST', '/tasks/{}/task_actions/complete'.format(task_id), data)

    async def still_running(self, task_id):
        """
        Check if the task should still be running according to the DB.

        Args:
            task_id (str): task_id of task
        """
        ret = await self.rest.request('GET', '/tasks/{}'.format(task_id))
        if (not ret) or 'status' not in ret or ret['status'] != 'processing':
            raise Exception('task should be stopped')

    async def task_error(self, task_id, dataset_id=None, stats={},
                         start_time=None, reason=None, resources=None,
                         site=None):
        """
        Tell the server about the error experienced

        Args:
            task_id (str): task_id of task
            dataset_id (str): (optional) dataset_id of task
            stats (dict): (optional) task statistics
            start_time (float): (optional) task start time in unix seconds
            reason (str): (optional) one-line summary of error
            resources (dict): (optional) task resource usage
            site (str): (optional) site the task is running at
        """
        iceprod_stats = {}
        try:
            hostname = functions.gethostname()
            domain = '.'.join(hostname.split('.')[-2:])
            if start_time:
                t = time.time() - start_time
            elif resources and 'time' in resources and resources['time']:
                t = int(resources['time']*3600)
            else:
                t = None
            iceprod_stats = {
                'task_id': task_id,
                'hostname': hostname,
                'domain': domain,
                'time_used': t,
                'task_stats': json.dumps(stats),
                'time': datetime.utcnow().isoformat(),
                'error_summary': reason if reason else '',
            }
            if dataset_id:
                iceprod_stats['dataset_id'] = dataset_id
            if resources:
                iceprod_stats['resources'] = resources
            if site:
                iceprod_stats['site'] = site
        except Exception:
            logging.warning('failed to collect error info', exc_info=True)

        try:
            await self.rest.request('POST', '/tasks/{}/task_stats'.format(task_id),
                                    iceprod_stats)
        except Exception:
            logging.warning('failed to post task_stats for %r', task_id, exc_info=True)

        data = {}
        if t:
            data['time_used'] = t
        if resources:
            data['resources'] = resources
        if site:
            data['site'] = site
        if reason:
            data['reason'] = reason
        await self.rest.request('POST', '/tasks/{}/task_actions/reset'.format(task_id), data)

    async def task_kill(self, task_id, dataset_id=None, resources=None,
                        reason=None, message=None, site=None):
        """
        Tell the server that we killed a task.

        Args:
            task_id (str): task_id of task
            dataset_id (str): (optional) dataset_id of task
            resources (dict): (optional) used resources
            reason (str): (optional) short summary for kill
            message (str): (optional) long message to replace log upload
            site (str): (optional) site the task is running at
        """
        if not reason:
            reason = 'killed'
        if not message:
            message = reason
        try:
            hostname = functions.gethostname()
            domain = '.'.join(hostname.split('.')[-2:])
            iceprod_stats = {
                'task_id': task_id,
                'hostname': hostname,
                'domain': domain,
                'time': datetime.utcnow().isoformat(),
                'error_summary': reason if reason else '',
            }
            if dataset_id:
                iceprod_stats['dataset_id'] = dataset_id
            if resources:
                iceprod_stats['resources'] = resources
            if site:
                iceprod_stats['site'] = site
        except Exception:
            logging.warning('failed to collect error info', exc_info=True)
            iceprod_stats = {}
        try:
            await self.rest.request('POST', '/tasks/{}/task_stats'.format(task_id),
                                    iceprod_stats)
        except Exception:
            logging.warning('failed to post task_stats for %r', task_id, exc_info=True)

        data = {}
        if resources and 'time' in resources and resources['time']:
            data['time_used'] = resources['time']*3600.
        if resources:
            data['resources'] = resources
        if site:
            data['site'] = site
        if reason:
            data['reason'] = reason
        else:
            data['data'] = 'task killed'
        await self.rest.request('POST', '/tasks/{}/task_actions/reset'.format(task_id), data)

        data = {'name': 'stdlog', 'task_id': task_id}
        if dataset_id:
            data['dataset_id'] = dataset_id
        if message:
            data['data'] = message
        elif reason:
            data['data'] = reason
        else:
            data['data'] = 'task killed'
        await self.rest.request('POST', '/logs', data)
        data['name'] = 'stdout'
        await self.rest.request('POST', '/logs', data)
        data['name'] = 'stderr'
        await self.rest.request('POST', '/logs', data)

    async def _upload_logfile(self, name, filename, task_id=None, dataset_id=None):
        """Upload a log file"""
        data = {'name': name}
        if task_id:
            data['task_id'] = task_id
        if dataset_id:
            data['dataset_id'] = dataset_id
        try:
            if filename.endswith('.gz'):
                with gzip.open(filename, 'r') as f:
                    data['data'] = f.read()
            else:
                with open(filename) as f:
                    data['data'] = f.read()
        except Exception as e:
            data['data'] = str(e)
        await self.rest.request('POST', '/logs', data)

    async def uploadLog(self, **kwargs):
        """Upload log file"""
        logging.getLogger().handlers[0].flush()
        filename = os.path.abspath(constants['stdlog'])
        if os.path.exists(filename):
            await self._upload_logfile('stdlog', filename, **kwargs)
        elif os.path.exists(filename+'.gz'):
            await self._upload_logfile('stdlog', filename+'.gz', **kwargs)
        else:
            raise Exception('cannot find stdlog')

    async def uploadErr(self, filename=None, **kwargs):
        """Upload stderr file"""
        if not filename:
            sys.stderr.flush()
            filename = os.path.abspath(constants['stderr'])
        await self._upload_logfile('stderr', filename, **kwargs)

    async def uploadOut(self, filename=None, **kwargs):
        """Upload stdout file"""
        if not filename:
            sys.stdout.flush()
            filename = os.path.abspath(constants['stdout'])
        await self._upload_logfile('stdout', filename, **kwargs)

    async def create_pilot(self, **kwargs):
        """
        Create an entry in the pilot table.

        Args:
            **kwargs: passed through to rest function
        Returns:
            str: pilot id
        """
        ret = await self.rest.request('POST', '/pilots', kwargs)
        return ret['result']

    async def update_pilot(self, pilot_id, **kwargs):
        """
        Update the pilot table.

        Args:
            pilot_id (str): pilot id
            **kwargs: passed through to rest function
        """
        await self.rest.request('PATCH', '/pilots/{}'.format(pilot_id), kwargs)

    async def delete_pilot(self, pilot_id, **kwargs):
        """
        Delete the pilot.

        Args:
            pilot_id (str): pilot id
        """
        await self.rest.request('DELETE', '/pilots/{}'.format(pilot_id))

    # --- synchronous versions to be used from a signal handler
    # --- or other non-async code

    def task_kill_sync(self, task_id, dataset_id=None, resources=None,
                       reason=None, message=None, site=None):
        """
        Tell the server that we killed a task (synchronous version).

        Args:
            task_id (str): task_id of task
            dataset_id (str): (optional) dataset_id of task
            resources (dict): (optional) used resources
            reason (str): (optional) short summary for kill
            message (str): (optional) long message to replace log upload
            site (str): (optional) site the task is running at
        """
        if not reason:
            reason = 'killed'
        if not message:
            message = reason
        try:
            hostname = functions.gethostname()
            domain = '.'.join(hostname.split('.')[-2:])
            iceprod_stats = {
                'task_id': task_id,
                'hostname': hostname,
                'domain': domain,
                'time': datetime.utcnow().isoformat(),
                'error_summary': reason if reason else '',
            }
            if dataset_id:
                iceprod_stats['dataset_id'] = dataset_id
            if resources:
                iceprod_stats['resources'] = resources
            if site:
                iceprod_stats['site'] = site
        except Exception:
            logging.warning('failed to collect error info', exc_info=True)
            iceprod_stats = {}
        try:
            self.rest.request_seq('POST', '/tasks/{}/task_stats'.format(task_id),
                                  iceprod_stats)
        except Exception:
            logging.warning('failed to post task_stats for %r', task_id, exc_info=True)

        data = {}
        if resources and 'time' in resources and resources['time']:
            data['time_used'] = resources['time']*3600.
        if resources:
            data['resources'] = resources
        if site:
            data['site'] = site
        if reason:
            data['reason'] = reason
        else:
            data['data'] = 'task killed'
        self.rest.request_seq('POST', '/tasks/{}/task_actions/reset'.format(task_id), data)

        data = {'name': 'stdlog', 'task_id': task_id}
        if dataset_id:
            data['dataset_id'] = dataset_id
        if message:
            data['data'] = message
        elif reason:
            data['data'] = reason
        else:
            data['data'] = 'task killed'
        self.rest.request_seq('POST', '/logs', data)
        data['name'] = 'stdout'
        self.rest.request_seq('POST', '/logs', data)
        data['name'] = 'stderr'
        self.rest.request_seq('POST', '/logs', data)

    def update_pilot_sync(self, pilot_id, **kwargs):
        """
        Update the pilot table (synchronous version).

        Args:
            pilot_id (str): pilot id
            **kwargs: passed through to rpc function
        """
        self.rest.request_seq('PATCH', '/pilots/{}'.format(pilot_id), kwargs)

    def delete_pilot_sync(self, pilot_id, **kwargs):
        """
        Delete the pilot (synchronous version).

        Args:
            pilot_id (str): pilot id
        """
        self.rest.request_seq('DELETE', '/pilots/{}'.format(pilot_id))



"""
Server
======

Run the iceprod server.
"""

import os
import sys
import logging
import importlib
import subprocess
from datetime import datetime, timedelta
import asyncio

from concurrent.futures import ThreadPoolExecutor

from iceprod.core.logger import set_log_level
from iceprod.server.config import IceProdConfig


logger = logging.getLogger('Server')


class Server(object):
    """
    The actual server.

    """
    def __init__(self, config_params=None, outfile=None, errfile=None):
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.cfg = IceProdConfig(override=config_params)
        self.modules = {}
        self.services = {
            'daemon': {
                'restart': self.restart,
                'reload': self.reload,
                'stop': self.stop,
                'kill': self.kill,
                'get_running_modules': lambda: self.modules.keys(),
            },
        }
        self.outfile = outfile
        self.errfile = errfile

        set_log_level(self.cfg['logging']['level'])

        for mod_name in self.cfg['modules']:
            if self.cfg['modules'][mod_name]:
                try:
                    m = importlib.import_module('iceprod.server.modules.'+mod_name)
                    mod = getattr(m, mod_name)(cfg=self.cfg,
                                               executor=self.executor,
                                               modules=self.services)
                    self.modules[mod_name] = mod
                    self.services[mod_name] = mod.service
                    mod.start()
                except Exception:
                    logger.critical('cannot start module', exc_info=True)
                    self.kill()
                    raise

    async def rotate_logs(self):
        current_date = datetime.utcnow()
        while self.outfile and self.errfile:
            if current_date.day != datetime.utcnow().day:
                # rotate files
                current_date = datetime.utcnow()
                if self.outfile:
                    roll_files(sys.stdout, self.outfile)
                if self.errfile:
                    roll_files(sys.stderr, self.errfile)
            await asyncio.sleep(3600)

    def start(self):
        asyncio.create_task(self.rotate_logs())

    def restart(self):
        env = os.environ.copy()
        extra_path = os.path.join(os.environ['I3PROD'],'bin')
        env['PATH'] = extra_path+(':'+env['PATH'] if 'PATH' in env else '')
        subprocess.Popen(['iceprod_server.py','restart'],
                         cwd=os.environ['I3PROD'], env=env)

    def reload(self):
        for m in self.modules.values():
            m.stop()
            m.start()

    def stop(self):
        for m in self.modules.values():
            m.stop()

    def kill(self):
        for m in self.modules.values():
            m.kill()


def roll_files(fd, filename, num_files=5):
    d = datetime.utcnow()
    ext = (d-timedelta(days=num_files-1)).strftime('%Y-%m-%d')
    newfile = f'{filename}.{ext}'
    if os.path.exists(newfile):  # delete last file
        os.remove(newfile)
    for i in range(num_files-2, 0, -1):
        ext = (d-timedelta(days=i)).strftime('%Y-%m-%d')
        oldfile = f'{filename}.{ext}'
        if os.path.exists(oldfile):
            os.rename(oldfile, newfile)
        newfile = oldfile
    if os.path.exists(filename):
        os.rename(filename, newfile)

    # redirect file descriptor
    newfd = open(filename, fd.mode)
    fd.flush()
    os.dup2(newfd.fileno(), fd.fileno())
    if fd != sys.stdout and fd != sys.stderr:
        fd.close()
    return newfd



import logging
import json
import uuid
from collections import defaultdict

import pymongo
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization, attr_auth
from iceprod.server.util import nowstr, dataset_statuses, dataset_status_sort

logger = logging.getLogger('rest.datasets')


def setup(handler_cfg):
    """
    Setup method for Dataset REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/datasets', MultiDatasetHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)', DatasetHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/description', DatasetDescriptionHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/status', DatasetStatusHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/priority', DatasetPriorityHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/jobs_submitted', DatasetJobsSubmittedHandler, handler_cfg),
            (r'/dataset_summaries/status', DatasetSummariesStatusHandler, handler_cfg),
        ],
        'database': 'datasets',
        'indexes': {
            'datasets': {
                'dataset_id_index': {'keys': 'dataset_id', 'unique': True},
            }
        },
    }


class MultiDatasetHandler(APIBase):
    """
    Handle multi-dataset requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    async def get(self):
        """
        Get a dict of datasets.

        Params (optional):
            status: | separated list of status filters
            groups: | separated list of groups to filter on
            users: | separated list of users to filter on
            keys: | separated list of keys to return for each dataset

        Returns:
            dict: {<dataset_id>: metadata}
        """
        query = {}
        status = self.get_argument('status', None)
        if status:
            query['status'] = {'$in': status.split('|')}
        groups = self.get_argument('groups', None)
        if groups:
            query['group'] = {'$in': groups.split('|')}
        users = self.get_argument('users', None)
        if users:
            query['username'] = {'$in': users.split('|')}

        projection = {'_id': False}
        keys = self.get_argument('keys', None)
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
        projection['dataset_id'] = True  # must be in projection

        ret = {}
        async for row in self.db.datasets.find(query, projection=projection):
            k = row['dataset_id']
            ret[k] = row
        self.write(ret)
        self.finish()

    @authorization(roles=['admin', 'user'])
    async def post(self):
        """
        Add a dataset.

        Body should contain all necessary fields for a dataset.
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'description': str,
            'jobs_submitted': int,
            'tasks_submitted': int,
            'tasks_per_job': int,
            'group': str,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key "{}" should be of type {}'.format(k, req_fields[k].__name__)
                raise tornado.web.HTTPError(400, reason=r)

        opt_fields = {
            'priority': float,
            'debug': bool,
            'jobs_immutable': bool,
            'status': str,
        }
        for k in opt_fields:
            if k in data and not isinstance(data[k], opt_fields[k]):
                r = 'key "{}" should be of type {}'.format(k, opt_fields[k].__name__)
                raise tornado.web.HTTPError(400, reason=r)

        bad_fields = set(data).difference(set(opt_fields).union(req_fields))
        if bad_fields:
            r = 'invalid keys found'
            raise tornado.web.HTTPError(400, reason=r)

        if data['jobs_submitted'] == 0 and data['tasks_per_job'] <= 0:
            r = '"tasks_per_job" must be > 0'
            raise tornado.web.HTTPError(400, reason=r)
        elif data['tasks_submitted'] != 0 and data['tasks_submitted'] / data['jobs_submitted'] != data['tasks_per_job']:
            r = '"tasks_per_job" does not match "tasks_submitted"/"jobs_submitted"'
            raise tornado.web.HTTPError(400, reason=r)

        # generate dataset number
        ret = await self.db.settings.find_one_and_update(
            {'name': 'dataset_num'},
            {'$inc': {'num': 1}},
            projection={'num': True, '_id': False},
            upsert=True,
            return_document=pymongo.ReturnDocument.AFTER)
        dataset_num = ret['num']

        # set some fields
        dataset_id = uuid.uuid1().hex
        data['dataset_id'] = dataset_id
        data['dataset'] = dataset_num
        if 'status' not in data:
            data['status'] = 'processing'
        data['start_date'] = nowstr()
        data['username'] = self.current_user
        if 'priority' not in data:
            data['priority'] = 0.5
        if 'debug' not in data:
            data['debug'] = False
        if 'jobs_immutable' not in data:
            data['jobs_immutable'] = False

        # insert
        ret = await self.db.datasets.insert_one(data)

        # set auth rules
        write_groups = list({'admin', data['group']}) if data['group'] != 'users' else ['admin']
        await self.set_attr_auth(
            'dataset_id',
            data['dataset_id'],
            read_groups=list({'admin', data['group'], 'users'}),
            write_groups=write_groups,
            read_users=[data['username']],
            write_users=[data['username']],
        )

        # make sure user prio is set
        try:
            await self.add_user(self.current_user)
        except pymongo.errors.DuplicateKeyError:
            # ignore already added users
            pass

        # return success
        self.set_status(201)
        self.set_header('Location', f'/datasets/{dataset_id}')
        self.write({'result': f'/datasets/{dataset_id}'})
        self.finish()


class DatasetHandler(APIBase):
    """
    Handle dataset requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get a dataset.

        Args:
            dataset_id (str): the dataset

        Returns:
            dict: dataset metadata
        """
        ret = await self.db.datasets.find_one({'dataset_id':dataset_id}, projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Dataset not found")
        else:
            self.write(ret)
            self.finish()


class DatasetDescriptionHandler(APIBase):
    """
    Handle dataset description updates.
    """
    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def put(self, dataset_id):
        """
        Set a dataset description.

        Args:
            dataset_id (str): the dataset

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'description' not in data:
            raise tornado.web.HTTPError(400, reason='missing description')
        elif not isinstance(data['description'],str):
            raise tornado.web.HTTPError(400, reason='bad description')

        ret = await self.db.datasets.find_one_and_update(
            {'dataset_id':dataset_id},
            {'$set':{'description': data['description']}},
            projection=['_id']
        )
        if not ret:
            self.send_error(404, reason="Dataset not found")
        else:
            self.write({})
            self.finish()


class DatasetStatusHandler(APIBase):
    """
    Handle dataset status updates.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def put(self, dataset_id):
        """
        Set a dataset status.

        Args:
            dataset_id (str): the dataset

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'status' not in data:
            raise tornado.web.HTTPError(400, reason='missing status')
        elif data['status'] not in dataset_statuses:
            raise tornado.web.HTTPError(400, reason='bad status')

        ret = await self.db.datasets.find_one_and_update(
            {'dataset_id':dataset_id},
            {'$set':{'status': data['status']}},
            projection=['_id']
        )
        if not ret:
            self.send_error(404, reason="Dataset not found")
        else:
            self.write({})
            self.finish()


class DatasetPriorityHandler(APIBase):
    """
    Handle dataset priority updates.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def put(self, dataset_id):
        """
        Set a dataset priority.

        Args:
            dataset_id (str): the dataset

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'priority' not in data:
            raise tornado.web.HTTPError(400, reason='missing priority')
        elif not isinstance(data['priority'], (int, float)):
            raise tornado.web.HTTPError(400, reason='priority is not a number')

        ret = await self.db.datasets.find_one_and_update(
            {'dataset_id':dataset_id},
            {'$set':{'priority': data['priority']}},
            projection=['_id']
        )
        if not ret:
            self.send_error(404, reason="Dataset not found")
        else:
            self.write({})
            self.finish()


class DatasetJobsSubmittedHandler(APIBase):
    """
    Handle dataset jobs_submitted updates.
    """
    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def put(self, dataset_id):
        """
        Set a dataset's jobs_submitted.

        Only allows increases, if the jobs_immutable flag is not set.

        Args:
            dataset_id (str): the dataset

        Json body:
            jobs_submitted (int): the number of jobs submitted

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'jobs_submitted' not in data:
            raise tornado.web.HTTPError(400, reason='missing jobs_submitted')
        try:
            jobs_submitted = int(data['jobs_submitted'])
        except Exception:
            raise tornado.web.HTTPError(400, reason='jobs_submitted is not an int')

        ret = await self.db.datasets.find_one({'dataset_id':dataset_id})
        if not ret:
            raise tornado.web.HTTPError(404, reason='Dataset not found')
        if ret['jobs_immutable']:
            raise tornado.web.HTTPError(400, reason='jobs_submitted is immutable')
        if ret['jobs_submitted'] > jobs_submitted:
            raise tornado.web.HTTPError(400, reason='jobs_submitted must be larger than before')
        if 'tasks_per_job' not in ret or ret['tasks_per_job'] <= 0:
            raise tornado.web.HTTPError(400, reason='tasks_per_job not valid')

        ret = await self.db.datasets.find_one_and_update(
            {'dataset_id':dataset_id},
            {'$set':{
                'jobs_submitted': jobs_submitted,
                'tasks_submitted': int(jobs_submitted*ret['tasks_per_job']),
            }},
            projection=['_id']
        )
        if not ret:
            self.send_error(404, reason="Dataset not found")
        else:
            self.write({})
            self.finish()


class DatasetSummariesStatusHandler(APIBase):
    """
    Handle dataset summary grouping by status.
    """
    @authorization(roles=['admin', 'user', 'system'])
    async def get(self):
        """
        Get the dataset summary for all datasets, group by status.

        Returns:
            dict: {<status>: [<dataset_id>,]}
        """
        cursor = self.db.datasets.find(projection={'_id':False,'status':True,'dataset_id':True})
        ret = defaultdict(list)
        async for row in cursor:
            ret[row['status']].append(row['dataset_id'])
        ret2 = {}
        for k in sorted(ret, key=dataset_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()



"""
A `recursive descent parser
<http://en.wikipedia.org/wiki/Recursive_descent_parser>`_ for the IceProd meta
language. Most commonly used in IceProd dataset configurations to refer to
other parts of the same configuration.
"""

from __future__ import absolute_import, division, print_function

import re
import random
import functools
import json
import builtins
import logging
import ast
import operator as op

from iceprod.core import dataclasses

logger = logging.getLogger('parser')


class safe_eval:
    # supported operators
    operators = {
        ast.Add: op.add,
        ast.Sub: op.sub,
        ast.Mult: op.mul,
        ast.Div: op.truediv,
        ast.FloorDiv: op.floordiv,
        ast.Mod: op.mod,
        ast.Pow: op.pow,
        ast.BitXor: op.xor,
        ast.Invert: op.invert,
        ast.Not: op.not_,
        ast.UAdd: op.pos,
        ast.USub: op.neg,
    }

    @classmethod
    def eval(cls,expr):
        """
        Safe evaluation of arithmatic operations using
        :mod:`Abstract Syntax Trees <ast>`.
        """
        return cls.__eval(ast.parse(expr).body[0].value)  # Module(body=[Expr(value=...)])

    @classmethod
    def __eval(cls,node):
        if isinstance(node, ast.Num):  # <number>
            return node.n
        elif isinstance(node, (ast.operator,ast.unaryop)):  # <operator>
            return cls.operators[type(node)]
        elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
            return cls.__eval(node.op)(cls.__eval(node.left), cls.__eval(node.right))
        elif isinstance(node, ast.UnaryOp):  # <operator> <right>
            return cls.__eval(node.op)(cls.__eval(node.operand))
        else:
            raise TypeError(node)


class GrammarException(Exception):
    pass


def getType(output):
    try:
        if isinstance(output,dataclasses.String) and not (output.startswith('"') and output.endswith('"')):
            try:
                output = json.loads(output.replace("'",'"'))
            except Exception:
                logging.debug('error formatting json', exc_info=True)
                if output.lower() == 'true':
                    output = True
                elif output.lower() == 'false':
                    output = False
                elif output.isdigit():
                    output = int(output)
                else:
                    output = float(output)
    except Exception:
        pass
    return output


def parse_ret_type(ret):
    if isinstance(ret, (list,dict)):
        return ret
    else:
        return str(ret)


tokens = ["word", "name", "starter", "scopeL", "scopeR", "bracketL", "bracketR"]


def scanner(data):
    """A lexical scanner, yielding token pairs"""
    word = ''
    escape = False
    for ch in data:
        if escape:
            word += ch
            escape = False
        elif ch == '\\':
            escape = True
        elif ch in '$()[]':
            if word:
                yield ('word', word)
                word = ''
            if ch == '$':
                yield ('starter', '$')
            elif ch == '(':
                yield ('scopeL', '(')
            elif ch == ')':
                yield ('scopeR', ')')
            elif ch == '[':
                yield ('bracketL', '[')
            elif ch == ']':
                yield ('bracketR', ']')
        else:
            word += ch
    if word:
        yield ('word', word)


def parser(data):
    """A syntactic parser, yielding syntactically accurate token pairs"""
    last_token = None
    last_word = ''
    nestings = 0
    stack = []
    for token,word in scanner(data):
        logger.debug('%s,%s,%s,%s,%d,%r',token,word,last_token,last_word,nestings,stack)
        if token == 'starter':
            if last_token == 'word':
                yield ("word", last_word)
            if last_token:
                stack.append(nestings)
                nestings = 0
                last_word = ''
            yield ("starter", '$')
            last_token = 'starter'
        elif token == 'scopeL':
            if last_token == 'starter':
                yield ('name', None)
                last_token = 'name'
            if last_token == 'name':
                yield (token,word)
                last_token = 'scopeL'
                last_word = ''
            else:
                # must be part of a broken word
                if last_token == 'word':
                    last_word += word
                else:
                    last_token = 'word'
                    last_word = word
                nestings += 1
        elif token == 'scopeR':
            if last_token == 'word':
                if nestings > 0:
                    last_word += word
                    nestings -= 1
                else:
                    yield ("word", last_word)
                    yield (token,word)
                    if stack:
                        nestings = stack.pop()
                    else:
                        nestings = 0
                    last_word = ''
            elif last_token == 'scopeL':
                yield ("word", '')
                yield (token,word)
                last_token = 'word'
                last_word = ''
            else:
                raise SyntaxError()
        elif token == 'bracketL':
            if last_token == 'word':
                if last_word:
                    # part of a broken word
                    last_word += '['
                    nestings += 1
                else:
                    yield ('bracketL', '[')
                    last_token = 'bracketL'
            else:
                raise SyntaxError()
        elif token == 'bracketR':
            if last_token == 'word':
                if nestings > 0:
                    last_word += word
                    nestings -= 1
                else:
                    yield ("word", last_word)
                    yield (token,word)
                    if stack:
                        nestings = stack.pop()
                    else:
                        nestings = 0
                    last_word = ''
            elif last_token == 'bracketL':
                yield ("word", '')
                yield (token,word)
                last_token = 'word'
                last_word = ''
            else:
                raise SyntaxError()
        elif token == 'word':
            if last_token == 'starter':
                yield ('name', word)
                last_token = 'name'
            elif last_token == 'word':
                last_word += word
            else:
                last_token = 'word'
                last_word = word
        else:
            # bad token
            raise SyntaxError()
    if last_word:
        yield ("word", last_word)
    if nestings or stack:
        raise SyntaxError()


class ExpParser:
    """
    Expression parsing class for parameter values.

    Grammar Definition::

        char     := any unicode character other than $()[]
        word     := char | char + word
        starter  := $
        scopeL   := (
        scopeR   := )
        bracketL := [
        bracketR := ]
        symbol   := starter | starter + word
        phrase   := symbol + scopeL + sentence + scopeR
        lookup   := word + bracketL + word + bracketR | phrase + bracketL + word + bracketR
        sentence := lookup | phrase | word | lookup + sentence | phrase + sentence | word + sentence

    Keywords:

    * steering : A parameter from :class:`iceprod.core.dataclasses.Steering`
    * system : A system value from :class:`iceprod.core.dataclasses.Steering`
    * args, options : An option value from :class:`iceprod.core.dataclasses.Job`
    * metadata : A value from :class:`iceprod.core.dataclasses.Dif` or
      :class:`iceprod.core.dataclasses.Plus`
    * eval : An arithmatic expression
    * sum, min, max, len : Apply a reduction to a sequence
    * choice : A random choice from a list of possibilites
    * sprintf : The sprintf string syntax

    Examples::

        $steering(my_parameter)
        $system(gpu_opts)
        $args(option1)
        $options(option1)
        $metadata(sensor_name)
        $eval(1+2)
        $choice(1,2,3,4)
        $sprintf("%04d",4)
    """
    def __init__(self):
        self.job = None
        self.env = None
        self.depth = 0
        # dict of keyword : function mappings
        self.keywords = {
            'steering' : self.steering_func,
            'system' : self.system_func,
            'environ' : self.environ_func,
            'args' : self.options_func,
            'options' : self.options_func,
            'metadata' : self.difplus_func,
            'eval' : self.eval_func,
            'choice' : self.choice_func,
            'sprintf' : self.sprintf_func
        }
        for reduction in 'sum', 'min', 'max', 'len':
            self.keywords[reduction] = functools.partial(self.reduce_func, getattr(builtins, reduction))

    def parse(self,input,job=None,env=None,depth=20):
        """
        Parse the input, expanding where possible.

        :param input: input string
        :param job: :class:`iceprod.core.dataclasses.Job`, optional
        :param env: env dictionary, optional
        :param depth: how deep to recursively parse
        :returns: expanded string
        """
        if depth < 1:
            logger.warning("recursion depth of parse exceeded")
            return input

        logger.debug("parse: %r",input)
        if not isinstance(input,dataclasses.String) or not input:
            # check for lists or dicts to recurse into
            if isinstance(input, list):
                logger.debug("recursing into list: %r", input)
                input = [self.parse(x, job=job, env=env, depth=depth-1) for x in input]
            elif isinstance(input, dict):
                logger.debug("recursing into dict: %r", input)
                input = {self.parse(x, job=job, env=env, depth=depth-1):self.parse(input[x],job=job,env=env,depth=depth-1) for x in input}
            return input

        # set job and env
        if job:
            self.job = job
        else:
            self.job = dataclasses.Job()
        if env:
            self.env = env
        else:
            self.env = {}
        self.depth = 0  # start at a depth of 0

        while True:
            # parse input
            stack = []
            try:
                for token,word in parser(input):
                    logger.debug('exp %s,%s,%r',token,word,stack)
                    if token in ('starter','name','word','scopeL','bracketL'):
                        stack.append((token,word))
                    elif token == 'scopeR':
                        # coelsce stack up to scopeL
                        word = ''
                        while stack and stack[-1][0] != 'scopeL':
                            word = stack.pop()[1] + word
                        stack.pop()  # remove scopeL
                        name = stack.pop()[1]
                        stack.pop()  # remove starter

                        # try evaluating this
                        try:
                            args = []
                            if name:
                                args.append(name)
                            args.append(word)
                            ret = self.process_phrase(*args)
                            if isinstance(ret, (list, dict)):
                                ret = json.dumps(ret)
                            stack.append(('word',str(ret)))
                        except GrammarException:
                            logger.debug('GrammarException')
                            stack.append(('word','$'+(name if name else '')+'('+word+')'))
                    elif token == 'bracketR':
                        # coelsce stack up to bracketL
                        word = ''
                        while stack and stack[-1][0] != 'bracketL':
                            word = stack.pop()[1] + word
                        stack.pop()  # remove bracketL

                        # try evaluating this
                        if word.endswith(']'):
                            # nested bracket, so recurse
                            word = self.parse(word, job=job, env=env, depth=depth-1)
                        try:
                            ret = self.process_phrase(word)
                        except GrammarException:
                            ret = word

                        # coelsce words
                        word = ''
                        while (stack and stack[-1][0] == 'word'
                               and (not word.startswith('['))
                               and (not word.startswith('{'))):
                            word = stack.pop()[1] + word

                        # now do list/dict index
                        try:
                            innerType = getType(ret)
                            ret = getType(word)[innerType]
                            stack.append(('word',str(ret)))
                        except Exception:
                            logger.debug('cannot eval: %s[%s]', word, ret,
                                         exc_info=True)
                            stack.append(('word',word+'['+ret+']'))
                    else:
                        raise SyntaxError()
            except Exception:
                logger.debug('SyntaxError', exc_info=True)
                output = getType(input)
            else:
                logger.debug('joining stack: %r', ''.join(s[1] for s in stack))
                output = getType(''.join(s[1] for s in stack))
                if isinstance(output,dataclasses.String) and output != input:
                    logger.debug('reprocessing output: %r', output)
                    input = output
                    continue
            break

        # check for lists or dicts to recurse into
        if isinstance(output, (list,dict)):
            output = self.parse(output, job=job, env=env, depth=depth)

        # return parsed output
        logger.debug('parser out: %r',output)
        return output

    def process_phrase(self,keyword,param=None):
        # search for keyword in special list
        ret = None
        if keyword in self.keywords and param is not None:
            try:
                ret = self.keywords[keyword](param)
            except GrammarException:
                pass
        if ret is None and param is None:
            # do general search for keyword
            if 'parameters' in self.env and keyword in self.env['parameters']:
                # search env params first
                ret = self.env['parameters'][keyword]
            elif keyword in self.job and not isinstance(self.job[keyword],dict):
                # search job second
                try:
                    ret = self.job[keyword]
                except Exception:
                    pass
            elif keyword in self.job['options']:  # search options third
                try:
                    ret = self.job['options'][keyword]
                except Exception:
                    pass
            elif self.job['steering'] and keyword in self.job['steering']['parameters']:
                # search job steering last
                ret = self.job['steering']['parameters'][keyword]

        if ret is None:
            raise GrammarException()
        return parse_ret_type(ret)

    def steering_func(self,param):
        """Find param in steering"""
        if self.job['steering'] and param in self.job['steering']['parameters']:
            return parse_ret_type(self.job['steering']['parameters'][param])
        else:
            raise GrammarException('steering:'+param)

    def system_func(self,param):
        """Find param in steering.system"""
        if self.job['steering'] and param in self.job['steering']['system']:
            return parse_ret_type(self.job['steering']['system'][param])
        else:
            raise GrammarException('system:'+param)

    def environ_func(self,param):
        """Find param in env["environment"]"""
        if 'environment' in self.env and param in self.env['environment']:
            return parse_ret_type(self.env['environment'][param])
        else:
            raise GrammarException('environ:'+param)

    def options_func(self,param):
        """Find param in options"""
        if param in self.job['options']:
            return parse_ret_type(self.job['options'][param])
        else:
            raise GrammarException('options:'+param)

    def difplus_func(self,param):
        """Find param in dif plus"""
        try:
            # try dif, then plus
            return parse_ret_type(self.job['difplus']['dif'][param])
        except Exception:
            try:
                return parse_ret_type(self.job['difplus']['plus'][param])
            except Exception:
                raise GrammarException('difplus:'+param)

    def choice_func(self,param):
        """Evaluate param as choice expression"""
        if not param:
            raise GrammarException('no choices available')
        try:
            if isinstance(param,(tuple,list)):
                return parse_ret_type(random.choice(param))
            else:
                return parse_ret_type(random.choice(param.split(',')))
        except Exception:
            raise GrammarException('not a valid choice')

    def eval_func(self,param):
        """Evaluate param as arithmetic expression"""
        bad = functools.reduce(lambda a, b: a or (b in param),('import','open','for','while','def','class','lambda'),False)
        if bad:
            raise GrammarException('Unsafe operator call')
        else:
            try:
                return parse_ret_type(safe_eval.eval(param))
            except Exception:
                raise GrammarException('Eval is not basic arithmetic')

    def reduce_func(self, func, param):
        try:
            return parse_ret_type(func(getType(param)))
        except Exception:
            raise GrammarException('Not a reducible sequence')

    def sprintf_func(self,param):
        """Evaluate param as sprintf.  param = arg_str, arg0, arg1, ... """
        # separate into format string and args
        strchar = param[0]
        if strchar in '\'"':
            pos = param.find(strchar,1)
            if pos < 0:
                raise GrammarException("Can't find closing quote for format string")
            fmt_str = param[1:pos]
        else:
            pos = param.find(',',0)
            if pos < 0:
                raise GrammarException("Can't find end of format string")
            fmt_str = param[0:pos]
        args = []
        pos = param.find(',',pos)
        while pos >= 0:
            pos2 = param.find(',',pos+1)
            if pos2 < 0:
                args.append(param[pos+1:])
                break
            else:
                args.append(param[pos+1:pos2])
                pos = pos2

        try:
            # cast args to correct type
            def cast_string(fstring,arg):
                """cast string to value according to formatting character"""
                if not fstring:
                    return arg
                if fstring[-1] in 'cs':
                    if arg[0] in '\'"':
                        return str(arg[1:-1])
                    else:
                        return str(arg)
                elif fstring[-1] == 'r':
                    return repr(arg)
                elif fstring[-1].lower() in 'xo':
                    return int(arg)
                elif fstring[-1].lower() in 'idufeg':
                    return float(arg)
                else:
                    raise GrammarException('Unable to cast %s using format %s'%(arg,fstring))

            fstrings = re.findall(r'\%[#0\- +]{0,1}[0-9]*\.{0,1}[0-9]*[csridufegExXo]',fmt_str)
            args = list(map(cast_string,fstrings,args))[0:len(args)]

            # do sprintf on fmt_str and args
            if fstrings:
                return fmt_str % tuple(args)
            else:
                return fmt_str
        except Exception as e:
            raise GrammarException(str(e))



import asyncio
import logging
import time
from datetime import datetime, timedelta

import pymongo

from ..server.util import nowstr, datetime2str
from .materialize import Materialize

logger = logging.getLogger('materialization_service')


class MaterializationService:
    """Materialization service."""
    def __init__(self, database, rest_client, cleanup_hours=6):
        self.db = database
        self.materialize = Materialize(rest_client)
        self.cleanup_hours = cleanup_hours

        self.start_time = time.time()
        self.last_run_time = None
        self.last_cleanup_time = None
        self.last_success_time = None

    async def run(self):
        """
        Run loop.
        """
        while True:
            ret = await self._run_once()
            if ret == 'sleep':
                logger.info('materialization service has nothing to do. sleeping')
                await asyncio.sleep(60)

    async def _run_once(self):
        ret = None
        try:
            self.last_run_time = time.time()
            now = nowstr()

            # periodically cleanup
            if (not self.last_cleanup_time) or self.last_run_time - self.last_cleanup_time > 3600*self.cleanup_hours:
                clean_time = datetime2str(datetime.utcfromtimestamp(self.last_run_time)-timedelta(hours=self.cleanup_hours))
                await self.db.materialization.delete_many({'status': {'$in': ['complete', 'error']}, 'modify_timestamp': {'$lt': clean_time}})
                await self.db.materialization.update_many({'status': 'processing', 'modify_timestamp': {'$lt': clean_time}}, {'$set': {'status': 'waiting', 'modify_timestamp': now}})
                self.last_cleanup_time = time.time()

            # get next materialization from DB
            ret = await self.db.materialization.find_one_and_update(
                {'status': 'waiting'},
                {'$set': {'status': 'processing', 'modify_timestamp': now}},
                projection={'_id':False},
                sort=[('modify_timestamp', 1)],
                return_document=pymongo.ReturnDocument.AFTER,
            )
            if not ret:
                return 'sleep'

            # run materialization
            logger.warning(f'running materialization request {ret["materialization_id"]}')
            kwargs = {}
            if 'dataset_id' in ret and ret['dataset_id']:
                kwargs['only_dataset'] = ret['dataset_id']
            if 'num' in ret and ret['num']:
                kwargs['num'] = ret['num']
            if 'set_status' in ret and ret['set_status']:
                kwargs['set_status'] = ret['set_status']
            await self.materialize.run_once(**kwargs)

            await self.db.materialization.update_one(
                {'materialization_id': ret['materialization_id']},
                {'$set': {'status': 'complete'}},
            )
            self.last_success_time = time.time()
        except Exception:
            logger.error('error running materialization', exc_info=True)
            if ret:
                await self.db.materialization.update_one(
                    {'materialization_id': ret['materialization_id']},
                    {'$set': {'status': 'error'}},
                )



"""
Clean up the pilots.
"""

import argparse
import asyncio
import logging
from datetime import datetime, timedelta

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.util import str2datetime

logger = logging.getLogger('pilot_monitor')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    time_limit = datetime.utcnow() - timedelta(days=14)

    async def reset_pilot(pilot_id):
        await rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))
    try:
        pilots = await rest_client.request('GET', '/pilots')
        for pilot in pilots.values():
            if 'last_update' not in pilot or str2datetime(pilot['last_update']) < time_limit:
                await reset_pilot(pilot['pilot_id'])

    except Exception:
        logger.error('error cleaning pilots', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""
A supercomputer plugin, specificially for
`Cedar <https://docs.computecanada.ca/wiki/Cedar>`_.
"""
import asyncio
from datetime import datetime,timedelta
from functools import partial
import getpass
import logging
import os
import stat

from .condor_direct import check_call, check_output, condor_direct

logger = logging.getLogger('plugin-cedar')


async def subprocess_ssh(host, args):
    """A subprocess call over ssh"""
    cmd = ['ssh', '-i', '~/.ssh/iceprod', host]
    if 'ICEPRODBASE' in os.environ:
        cmd.append(f'{os.environ["ICEPRODBASE"]}/env-shell.sh')
    cmd += args
    logger.info('subprocess_ssh: %r', cmd)
    p = await asyncio.create_subprocess_exec(*cmd)
    await asyncio.wait_for(p.wait(), timeout=20*60)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p


class supercomp_cedar(condor_direct):
    """Plugin Overrides for Cedar"""

    batch_site = 'Cedar'
    batch_outfile = 'slurm.out'

    # Cedar queue requirements
    batch_resources = {
        'cpu': 1,
        'gpu': 0,
        'memory': 3.75,
        'disk': 48.,
        'time': 20.,
        'os': 'RHEL_7_x86_64',
    }

    async def get_hold_reason(self, submit_dir, resources=None):
        """Search for a hold reason in the slurm stderr"""
        reason = None
        filename = os.path.join(submit_dir, 'slurm.err')
        if os.path.exists(filename):
            with open(filename) as f:
                for line in f:
                    line = line.strip().lower()
                    if line.startswith('slurmstepd: error:'):
                        resource_type = None
                        val = 0
                        if 'oom-kill' in line:
                            resource_type = 'memory'
                            with open(os.path.join(submit_dir, 'submit.sh')) as f2:
                                for line2 in f2:
                                    if line2.startswith('export NUM_MEMORY='):
                                        try:
                                            val = float(line2.split('=')[-1].strip())
                                        except Exception:
                                            pass
                                        break
                        elif 'due to time limit' in line:
                            resource_type = 'time'
                            with open(os.path.join(submit_dir, 'submit.sh')) as f2:
                                for line2 in f2:
                                    if line2.startswith('export NUM_TIME='):
                                        try:
                                            val = float(line2.split('=')[-1].strip())
                                        except Exception:
                                            pass
                                        break
                        if resource_type:
                            reason = f'Resource overusage for {resource_type}: {val}'
                            if val:
                                resources[resource_type] = val
                            break
        return reason

    async def download_input(self, task):
        """
        Download input files for task.

        Args:
            task (dict): task info
        """
        self.x509proxy.get_proxy()
        try:
            await check_call(
                'python', '-m', 'iceprod.core.data_transfer', '-f',
                os.path.join(task['submit_dir'],'task.cfg'),
                '-d', task['submit_dir'],
                'input'
            )
        except Exception:
            logger.info('failed to download input', exc_info=True)
            raise Exception(f'download input failed for {task["dataset_id"]}.{task["task_id"]}')

    async def upload_output(self, task):
        """
        Upload output files for task.

        Args:
            task (dict): task info
        """
        self.x509proxy.get_proxy()
        try:
            await check_call(
                'python', '-m', 'iceprod.core.data_transfer', '-f',
                os.path.join(task['submit_dir'],'task.cfg'),
                '-d', task['submit_dir'],
                'output'
            )
        except Exception:
            logger.info('failed to upload output', exc_info=True)
            raise Exception(f'upload output failed for {task["dataset_id"]}.{task["task_id"]}')

    async def generate_submit_file(self, task, cfg=None, passkey=None, filelist=None):
        """Generate queueing system submit file for task in dir."""
        args = self.get_submit_args(task,cfg=cfg)
        args.extend(['--offline',])

        # write the submit file
        submit_file = os.path.join(task['submit_dir'],'submit.sh')
        with open(submit_file,'w') as f:
            p = partial(print,sep='',file=f)
            p('#!/bin/bash')
            p('#SBATCH --account=rpp-kenclark')
            p('#SBATCH --output={}'.format(os.path.join(task['submit_dir'],'slurm.out')))
            p('#SBATCH --error={}'.format(os.path.join(task['submit_dir'],'slurm.err')))
            p(f'#SBATCH --chdir={task["submit_dir"]}')
            p('#SBATCH --ntasks=1')
            p('#SBATCH --export=NONE')
            p('#SBATCH --mail-type=NONE')
            p('#SBATCH --job-name=iceprod_{}'.format(os.path.basename(task['submit_dir'])))

            # handle resource requests
            if 'requirements' in task:
                if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                    p(f'#SBATCH --cpus-per-task={task["requirements"]["cpu"]}')
                if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                    p(f'#SBATCH --gres=gpu:{task["requirements"]["gpu"]}')
                if 'memory' in task['requirements'] and task['requirements']['memory']:
                    p('#SBATCH --mem={}M'.format(int(task['requirements']['memory']*1000)))
                # we don't currently use the local disk, just the global scratch
                # if 'disk' in task['requirements'] and task['requirements']['disk']:
                #     p('#SBATCH --tmp={}M'.format(int(task['requirements']['disk']*1000)))
                if 'time' in task['requirements'] and task['requirements']['time']:
                    p('#SBATCH --time={}'.format(int(task['requirements']['time']*60)))

            # get batchopts
            for b in self.queue_cfg['batchopts']:
                p(b+'='+self.queue_cfg['batchopts'][b])

            # make resources explicit in env
            if 'requirements' in task:
                if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                    p(f'export NUM_CPUS={task["requirements"]["cpu"]}')
                if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                    p(f'export NUM_GPUS={task["requirements"]["gpu"]}')
                if 'memory' in task['requirements'] and task['requirements']['memory']:
                    p(f'export NUM_MEMORY={task["requirements"]["memory"]}')
                if 'disk' in task['requirements'] and task['requirements']['disk']:
                    p(f'export NUM_DISK={task["requirements"]["disk"]}')
                if 'time' in task['requirements'] and task['requirements']['time']:
                    p(f'export NUM_TIME={task["requirements"]["time"]}')

            p('module load singularity/3.2')
            p('/opt/software/singularity-3.2/bin/singularity exec --nv --cleanenv -C', end=' ')
            p(f'-B /tmp -B /cvmfs -B /scratch -B /home --pwd {task["submit_dir"]}', end=' ')
            p('/cvmfs/singularity.opensciencegrid.org/opensciencegrid/osgvo-el7-cuda10:latest', end=' ')
            p('{} {}'.format(os.path.join(task['submit_dir'],'loader.sh'), ' '.join(args)))

        # make it executable
        st = os.stat(submit_file)
        os.chmod(submit_file, st.st_mode | stat.S_IEXEC)

    async def submit(self,task):
        """Submit task to queueing system."""
        cmd = ['sbatch','submit.sh']
        ret = await check_output(*cmd, cwd=task['submit_dir'])
        grid_queue_id = ''
        for line in ret.split('\n'):
            if 'Submitted batch job' in line:
                grid_queue_id = line.strip().rsplit(' ',1)[-1]
                break
        else:
            raise Exception('did not get a grid_queue_id')
        task['grid_queue_id'] = grid_queue_id

    async def get_grid_status(self):
        """
        Get all tasks running on the queue system.

        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        cmd = ['squeue', '-u', getpass.getuser(), '-h', '-o', '%A %t %j %o']
        out = await check_output(*cmd)
        ret = {}
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,name,cmd = line.split()
            if not name.startswith('iceprod'):
                continue
            if status == 'PD':
                status = 'queued'
            elif status == 'R':
                status = 'processing'
            elif status == 'CD':
                status = 'completed'
            else:
                status = 'error'
            ret[gid] = {'status': status, 'submit_dir': os.path.dirname(cmd),
                        'site': self.site}
        return ret

    async def get_grid_completions(self):
        """
        Get completions in the last 4 days.

        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        date = (datetime.now()-timedelta(days=4)).isoformat().split('.',1)[0]
        cmd = ['sacct', '-u', getpass.getuser(), '-n', '-P', '-S', date, '-o', 'JobIDRaw,State,JobName,ExitCode,Workdir']
        out = await check_output(*cmd)
        ret = {}
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,name,exit_code,workdir = line.strip().split('|')
            if status in ('PENDING', 'RESIZING', 'REQUEUED', 'RUNNING') or not name.startswith('iceprod'):
                continue
            if status == 'COMPLETED' and exit_code == '0:0':
                status = 'ok'
            else:
                status = 'error'
            ret[gid] = {'status': status, 'submit_dir': workdir, 'site': self.site}
        return ret

    async def remove(self,tasks):
        """Remove tasks from queueing system."""
        if tasks:
            cmd = ['scancel']+list(tasks)
            await check_call(*cmd)



"""
The core execution functions for running on a node.
These are all called from :any:`iceprod.core.i3exec`.

The fundamental design of the core is to run a task composed of trays and
modules.  The general heirarchy looks like::

    task
    |
    |- tray1
       |
       |- module1
       |
       |- module2
    |
    |- tray2
       |
       |- module3
       |
       |- module4

Parameters can be defined at every level, and each level is treated as a
scope (such that inner scopes inherit from outer scopes).  This is
accomplished via an internal evironment for each scope.
"""

from __future__ import absolute_import, division, print_function

import sys
import os
import stat
import time
import glob
import copy
from datetime import datetime
import asyncio

try:
    import cPickle as pickle
except Exception:
    import pickle

import logging


from iceprod.core import constants
from iceprod.core import dataclasses
from iceprod.core import util
from iceprod.core import functions
from iceprod.core.resources import Resources
import iceprod.core.parser
from iceprod.core.jsonUtil import json_encode,json_decode


class Config:
    """Contain the configuration and related methods"""
    def __init__(self, config=None, parser=None, rpc=None, logger=None):
        self.config = config if config else dataclasses.Job()
        self.parser = parser if parser else iceprod.core.parser.ExpParser()
        self.rpc = rpc
        self.logger = logger if logger else logging

    def parseValue(self, value, env={}):
        """
        Parse a value from the available env and global config.

        Uses the :class:`Meta Parser <iceprod.core.parser>` on any string value.
        Pass-through for any other object.

        :param value: The value to parse
        :param env: The environment to use, optional
        :returns: The parsed value
        """
        if isinstance(value,dataclasses.String):
            self.logger.debug('parse before:%r| env=%r',value,env)
            value = self.parser.parse(value,self.config,env)
            if isinstance(value,dataclasses.String):
                value = os.path.expandvars(value)
            self.logger.debug('parse after:%r',value)
        return value

    def parseObject(self,obj,env):
        """Recursively parse a dict or list"""
        if isinstance(obj,dataclasses.String):
            return self.parseValue(obj,env)
        elif isinstance(obj,(list,tuple)):
            return [self.parseObject(v,env) for v in obj]
        elif isinstance(obj,dict):
            ret = copy.copy(obj)  # in case it's a subclass of dict, like dataclasses
            for k in obj:
                ret[k] = self.parseObject(obj[k],env)
            return ret
        else:
            return obj


class SetupEnv:
    """
    The internal environment (env) is a dictionary composed of several objects:

    parameters
        Parameters are defined directly as an object, or as a string pointing
        to another object.  They can use the IceProd meta-language to be
        defined in relation to other parameters specified in inherited
        scopes, or as eval or sprinf functions.

    resources
        \\

    data
        Resources and data are similar in that they handle extra files that
        modules may create or use.  The difference is that resources are only
        for reading, such as pre-built lookup tables, while data can be input
        and/or output. Compression can be automatically handled by IceProd.
        Both resources and data are defined in the environment as strings to
        their file location.

    classes
        This is where external software gets added.  The software can be an
        already downloaded resource or just a url to download.  All python
        files get added to the python path and binary libraries get symlinked
        into a directory on the LD_LIBRARY_PATH.  Note that if there is more
        than one copy of the same shared library file, only the most recent
        one is in scope.  Classes are defined in the environment as strings
        to their file location.

    deletions
        These are files that should be deleted when the scope ends.

    uploads
        These are files that should be uploaded when the scope ends.
        Mostly Data objects that are used as output.

    shell environment
        An environment to reset to when exiting the context manager.

    To keep the scope correct a new dictionary is created for every level, then
    the inheritable objects are shallow copied (to 1 level) into the new env.
    The deletions are not inheritable (start empty for each scope), and the shell
    environment is set at whatever the previous scope currently has.

    Args:
        cfg (:py:class:`Config`): Config object
        obj (dict): A dict-like object from :py:mod:`iceprod.core.dataclasses`
                    such as :py:class:`iceprod.core.dataclasses.Steering`.
        oldenv (dict): (optional) env that we are running inside
    """
    def __init__(self, cfg, obj, oldenv={}, logger=None):
        self.cfg = cfg
        self.obj = obj
        self.oldenv = oldenv
        self.env = {}
        self.logger = logger if logger else logging

        # validation of input
        if not self.obj:
            raise util.NoncriticalError('object to load environment from is empty')
        if isinstance(self.obj, dataclasses.Steering) and not self.obj.valid():
            raise Exception('object is not valid Steering')

    async def __aenter__(self):
        try:
            # attempt to do depth=2 copying
            for key in self.oldenv:
                if key not in ('deletions','uploads','environment','pythonpath','stats'):
                    self.env[key] = copy.copy(self.oldenv[key])

            # make sure things for this env are clear (don't inherit)
            self.env['deletions'] = []
            self.env['uploads'] = []

            # get clear environment variables
            self.env['environment'] = os.environ.copy()
            self.env['pythonpath'] = copy.copy(sys.path)

            # inherit statistics
            if 'stats' in self.oldenv:
                self.env['stats'] = self.oldenv['stats']
            else:
                self.env['stats'] = {'upload':[], 'download':[], 'tasks':[]}

            # copy parameters
            if 'parameters' not in self.env:
                self.env['parameters'] = {}
            if 'parameters' in self.obj:
                # copy new parameters to env first so local referrals work
                self.env['parameters'].update(self.obj['parameters'])
                # parse parameter values and update if necessary
                for p in self.obj['parameters']:
                    newval = self.cfg.parseValue(self.obj['parameters'][p], self.env)
                    if newval != self.obj['parameters'][p]:
                        self.env['parameters'][p] = newval

            if 'resources' not in self.env:
                self.env['resources'] = {}
            if 'resources' in self.obj:
                # download resources
                for resource in self.obj['resources']:
                    await downloadResource(self.env, self.cfg.parseObject(resource, self.env), logger=self.logger)

            if 'data' not in self.env:
                self.env['data'] = {}
            input_files = self.cfg.config['options']['input'].split() if 'input' in self.cfg.config['options'] else []
            output_files = self.cfg.config['options']['output'].split() if 'output' in self.cfg.config['options'] else []
            if 'data' in self.obj:
                # download data
                for data in self.obj['data']:
                    d = self.cfg.parseObject(data, self.env)
                    if d['movement'] in ('input','both'):
                        await downloadData(self.env, d, logger=self.logger)
                        if 'local' in d and d['local']:
                            input_files.append(d['local'])
                        elif 'remote' in d and d['remote']:
                            input_files.append(os.path.basename(d['remote']))
                    if d['movement'] in ('output','both'):
                        self.env['uploads'].append(d)
                        if 'local' in d and d['local']:
                            output_files.append(d['local'])
                        elif 'remote' in d and d['remote']:
                            output_files.append(os.path.basename(d['remote']))
            # add input and output to parseable options
            self.cfg.config['options']['input'] = ' '.join(input_files)
            self.cfg.config['options']['output'] = ' '.join(output_files)
            logging.info('input: %r', self.cfg.config['options']['input'])
            logging.info('output: %r', self.cfg.config['options']['output'])

            if 'classes' not in self.env:
                self.env['classes'] = {}
            if 'classes' in self.obj:
                # set up classes
                for c in self.obj['classes']:
                    await setupClass(self.env, self.cfg.parseObject(c, self.env), logger=self.logger)

        except util.NoncriticalError:
            self.logger.warning('Noncritical error when setting up environment', exc_info=True)
        except Exception:
            self.logger.critical('Serious error when setting up environment', exc_info=True)
            raise

        return self.env

    async def __aexit__(self, exc_type, exc, tb):
        try:
            if not exc_type:
                # upload data if there was no exception
                if 'uploads' in self.env and (
                        'offline' not in self.cfg.config['options']
                        or (not self.cfg.config['options']['offline'])
                        or (
                            self.cfg.config['options']['offline']
                            and 'offline_transfer' in self.cfg.config['options']
                            and self.cfg.config['options']['offline_transfer']
                        )):
                    for d in self.env['uploads']:
                        await uploadData(self.env, d, logger=self.logger)
        finally:
            # delete any files
            if 'deletions' in self.env and len(self.env['deletions']) > 0:
                for f in reversed(self.env['deletions']):
                    try:
                        os.remove(f)
                        # base = os.path.basename(f)
                    except OSError as e:
                        self.logger.error('failed to delete file %s - %s',(str(f),str(e)))
                        if ('options' in self.env and
                                'debug' in self.env['options'] and
                                self.env['options']['debug']):
                            raise

            # reset environment
            if 'environment' in self.env:
                for e in list(os.environ.keys()):
                    if e not in self.env['environment']:
                        del os.environ[e]
                for e in self.env['environment'].keys():
                    os.environ[e] = self.env['environment'][e]


async def downloadResource(env, resource, remote_base=None,
                           local_base=None, checksum=None, logger=None):
    if not logger:
        logger = logging
    """Download a resource and put location in the env"""
    if not remote_base:
        remote_base = env['options']['resource_url']
    if not resource['remote'] and not resource['local']:
        raise Exception('need to specify either local or remote')
    if not resource['remote']:
        url = os.path.join(remote_base, resource['local'])
    elif functions.isurl(resource['remote']):
        url = resource['remote']
    else:
        url = os.path.join(remote_base,resource['remote'])

    execute = resource.do_transfer()
    if execute is False:
        logger.info('not transferring file %s', url)
        return

    if not local_base:
        if 'subprocess_dir' in env['options']:
            local_base = env['options']['subprocess_dir']
        else:
            local_base = os.getcwd()
    if not resource['local']:
        resource['local'] = os.path.basename(resource['remote'])
    local = os.path.join(local_base,resource['local'])
    if 'files' not in env:
        env['files'] = {}
    if not os.path.exists(os.path.dirname(local)):
        os.makedirs(os.path.dirname(local))

    # get resource
    if resource['local'] in env['files']:
        logger.info('resource %s already exists in env, so skip download and compression',resource['local'])
        return
    elif os.path.exists(local):
        logger.info('resource %s already exists as file, so skip download',resource['local'])
    else:
        # download resource
        download_options = {}
        if 'options' in env and 'username' in env['options']:
            download_options['username'] = env['options']['username']
        if 'options' in env and 'password' in env['options']:
            download_options['password'] = env['options']['password']
        if 'options' in env and 'ssl' in env['options'] and env['options']['ssl']:
            download_options.update(env['options']['ssl'])
        if 'options' in env and 'credentials' in env['options']:
            for base_url in env['options']['credentials']:
                if resource['remote'].startswith(base_url):
                    logger.info('using credential for %s', base_url)
                    cred_file = os.path.join(env['options']['credentials_dir'], env['options']['credentials'][base_url])
                    try:
                        with open(cred_file) as f:
                            token = f.read()
                    except Exception:
                        logger.critical('failed to load credential at %s', cred_file)
                        raise Exception('failed to download {} to {}'.format(url, local))
                    download_options['token'] = token
                    break
        failed = False
        try:
            start_time = time.time()
            await functions.download(url, local, options=download_options)
            if not os.path.exists(local):
                raise Exception('file does not exist')
            if checksum:
                # check the checksum
                cksm = functions.sha512sum(local)
                if cksm != checksum:
                    raise Exception('checksum validation failed')
        except Exception:
            if execute is False or execute == 'maybe':
                logger.info('not transferring file %s', url)
                return
            failed = True
            logger.critical('failed to download %s to %s', url, local, exc_info=True)
            raise Exception('failed to download {} to {}'.format(url, local))
        finally:
            stats = {
                'name': url,
                'error': failed,
                'now': datetime.utcnow().isoformat(),
                'duration': time.time()-start_time,
            }
            if (not failed) and os.path.exists(local):
                stats['size'] = os.path.getsize(local)
                stats['rate_MBps'] = stats['size']/1000/1000/stats['duration']

            if 'stats' in env and (execute is True or (execute == 'maybe' and not failed)):
                if 'download' not in env['stats']:
                    env['stats']['download'] = []
                env['stats']['download'].append(stats)

            if (not failed) and 'data_movement_stats' in env['options'] and env['options']['data_movement_stats']:
                print(f'{stats["now"]} Data movement stats: input {stats["duration"]:.3f} {stats["size"]:.0f} {stats["name"]}')

    # check compression
    if resource['compression'] and (functions.iscompressed(url) or functions.istarred(url)):
        # uncompress file
        files = functions.uncompress(local)
        # add uncompressed file(s) to env
        env['files'][resource['local']] = files
    else:
        # add file to env
        env['files'][resource['local']] = local
    logger.warning('resource %s added to env',resource['local'])


async def downloadData(env, data, logger=None):
    """Download data and put location in the env"""
    if not logger:
        logger = logging
    remote_base = data.storage_location(env)
    if 'options' in env and 'subprocess_dir' in env['options']:
        local_base = env['options']['subprocess_dir']
    else:
        local_base = os.getcwd()

    execute = data.do_transfer()
    checksum = None
    if execute is not False:
        try:
            filecatalog = data.filecatalog(env)
            path, checksum = filecatalog.get(data['local'])
        except Exception:
            # no filecatalog available
            pass
    await downloadResource(env, data, remote_base, local_base,
                           checksum=checksum, logger=logger)


async def uploadData(env, data, logger=None):
    """Upload data"""
    if not logger:
        logger = logging
    remote_base = data.storage_location(env)
    if 'options' in env and 'subprocess_dir' in env['options']:
        local_base = env['options']['subprocess_dir']
    else:
        local_base = os.getcwd()
    if (not data['remote']) and not data['local']:
        raise Exception('need either remote or local defined')
    if not data['remote']:
        url = os.path.join(remote_base, data['local'])
    elif not functions.isurl(data['remote']):
        url = os.path.join(remote_base, data['remote'])
    else:
        url = data['remote']

    if not data['local']:
        data['local'] = os.path.basename(data['remote'])
    local = os.path.join(local_base, data['local'])

    execute = data.do_transfer()
    exists = os.path.exists(local)
    if execute is False or (execute == 'maybe' and not exists):
        logger.info('not transferring file %s', local)
        return
    elif not exists:
        raise Exception('file {} does not exist'.format(local))

    # check compression
    if data['compression']:
        # get compression type, if specified
        if ((functions.iscompressed(url) or functions.istarred(url)) and
                not (functions.iscompressed(local) or functions.istarred(local))):
            # url has compression on it, so use that
            if '.tar.' in url:
                c = '.'.join(url.rsplit('.',2)[-2:])
            else:
                c = url.rsplit('.',1)[-1]
            try:
                local = functions.compress(local,c)
            except Exception:
                logger.warning('cannot compress file %s to %s', local, c)
                raise

    # upload file
    upload_options = {}
    if 'options' in env and 'username' in env['options']:
        upload_options['username'] = env['options']['username']
    if 'options' in env and 'password' in env['options']:
        upload_options['password'] = env['options']['password']
    if 'options' in env and 'ssl' in env['options'] and env['options']['ssl']:
        upload_options.update(env['options']['ssl'])
    if 'options' in env and 'credentials' in env['options']:
        for base_url in env['options']['credentials']:
            if data['remote'].startswith(base_url):
                logger.info('using credential for %s', base_url)
                cred_file = os.path.join(env['options']['credentials_dir'], env['options']['credentials'][base_url])
                try:
                    with open(cred_file) as f:
                        token = f.read()
                except Exception:
                    logger.critical('failed to load credential at %s', cred_file)
                    raise Exception('failed to upload {} to {}'.format(url, local))
                upload_options['token'] = token
                break
    do_checksum = True
    if 'options' in env and 'upload_checksum' in env['options']:
        do_checksum = env['options']['upload_checksum']
    failed = False
    try:
        start_time = time.time()
        await functions.upload(local, url, checksum=do_checksum, options=upload_options)
    except Exception:
        failed = True
        logger.critical('failed to upload %s to %s', local, url, exc_info=True)
        raise Exception('failed to upload {} to {}'.format(local, url))
    finally:
        stats = {
            'name': url,
            'error': failed,
            'now': datetime.utcnow().isoformat(),
            'duration': time.time()-start_time,
        }
        if not failed:
            stats['size'] = os.path.getsize(local)
            stats['rate_MBps'] = stats['size']/1000/1000/stats['duration']
        if 'stats' in env:
            env['stats']['upload'].append(stats)

        if (not failed) and 'data_movement_stats' in env['options'] and env['options']['data_movement_stats']:
            print(f'{stats["now"]} Data movement stats: output {stats["duration"]:.3f} {stats["size"]:.0f} {stats["name"]}')

    # if successful, add to filecatalog
    try:
        filecatalog = data.filecatalog(env)
    except Exception:
        pass  # no filecatalog available
    else:
        try:
            cksm = functions.sha512sum(local)
            metadata = {
                'file_size': stats['size'],
                'create_date': stats['now'],
                'modify_date': stats['now'],
                'data_type': 'simulation',
                'transfer_duration': stats['duration'],
                'transfer_MBps': stats['rate_MBps'],
            }
            options = ('dataset','dataset_id','task_id','task','job','debug')
            metadata.update({env['options'][k] for k in options if k in env['options']})
            filecatalog.add(data['local'], url, cksm, metadata)
        except Exception:
            logger.warning('failed to add %r to filecatalog', url, exc_info=True)


async def setupClass(env, class_obj, logger=None):
    """Set up a class for use in modules, and put it in the env"""
    if not logger:
        logger = logging
    if 'classes' not in env:
        env['classes'] = {}
    if not class_obj:
        raise Exception('Class is not defined')
    loaded = False
    if class_obj['name'] in env['classes']:
        # class already loaded, so leave it alone
        logger.info('class %s already loaded',class_obj['name'])
    elif class_obj['resource_name']:
        # class is downloaded as a resource
        if 'files' not in env or class_obj['resource_name'] not in env['files']:
            logger.error('resource %s for class %s does not exist',
                         class_obj['resource_name'],class_obj['name'])
        else:
            local = env['files'][class_obj['resource_name']]
            if not isinstance(local,dataclasses.String):
                local = local[0]
            if class_obj['src'] and os.path.exists(os.path.join(local,class_obj['src'])):
                # treat src as a path inside the resource
                local = os.path.join(local,class_obj['src'])
            loaded = True
    else:
        # get url of class
        i = 0
        while True:
            url = class_obj['src']
            if url and functions.isurl(url):
                i = 10  # skip repeat download attempts
            else:
                if i == 0:
                    # first, look in resources
                    if 'options' in env and 'resource_url' in env['options']:
                        url = os.path.join(env['options']['resource_url'],class_obj['src'])
                    else:
                        url = os.path.join('http://prod-exe.icecube.wisc.edu/',class_obj['src'])
                elif i == 1:
                    # then, look in regular svn
                    if 'options' in env and 'svn_repository' in env['options']:
                        url = os.path.join(env['options']['svn_repository'],class_obj['src'])
                    else:
                        url = os.path.join('http://code.icecube.wisc.edu/svn/projects/',class_obj['src'])
                else:
                    raise util.NoncriticalError('Cannot find class %s because of bad src url'%class_obj['name'])

            if 'options' in env and 'local_temp' in env['options']:
                local_temp = env['options']['local_temp']
            else:
                local_temp = os.path.join(os.getcwd(),'classes')
                env['options']['local_temp'] = local_temp
            if not os.path.exists(local_temp):
                os.makedirs(local_temp)
            if 'PYTHONPATH' in os.environ and local_temp not in os.environ['PYTHONPATH']:
                os.environ['PYTHONPATH'] += ':'+local_temp
            elif 'PYTHONPATH' not in os.environ:
                os.environ['PYTHONPATH'] = local_temp

            local = os.path.join(local_temp,class_obj['name'].replace(' ','_'))

            download_options = {}
            if 'options' in env and 'username' in env['options']:
                download_options['username'] = env['options']['username']
            if 'options' in env and 'password' in env['options']:
                download_options['password'] = env['options']['password']
            if 'options' in env and 'ssl' in env['options'] and env['options']['ssl']:
                download_options.update(env['options']['ssl'])

            # download class
            logger.warning('attempting to download class %s to %s',url,local_temp)
            try:
                download_local = await functions.download(url, local_temp, options=download_options)
            except Exception:
                logger.info('download failed, {} attempts left'.format(i), exc_info=True)
                if i < 10:
                    i += 1
                    continue  # retry with different url
                raise
            if not os.path.exists(download_local):
                raise Exception('failed to download {} to {}'.format(url, local))
            if functions.iscompressed(download_local) or functions.istarred(download_local):
                files = functions.uncompress(download_local, out_dir=local_temp)
                # check if we extracted a tarfile
                if isinstance(files,dataclasses.String):
                    local = files
                elif isinstance(files,list):
                    dirname = os.path.commonprefix(files)
                    if dirname:
                        dirname = os.path.join(local_temp, dirname.split(os.path.sep)[0])
                    else:
                        dirname = local_temp
                    logger.info('looking up tarball at %r', dirname)
                    if os.path.isdir(dirname):
                        logger.info('rename %r to %r', local, dirname)
                        local = dirname
                else:
                    logger.warning('files is strange datatype: %r', type(files))
            elif local != download_local:
                logger.info('rename %r to %r', download_local, local)
                os.rename(download_local, local)
            loaded = True
            break

    if loaded:
        # add to env
        env['classes'][class_obj['name']] = local
        logger.warning('class %s loaded at %r',class_obj['name'],local)

        # add binary libraries to the LD_LIBRARY_PATH
        def ldpath(root,f=None):
            root = os.path.abspath(root)

            def islib(f):
                return f[-3:] == '.so' or '.so.' in f or f[-2:] == '.a' or '.a.' in f
            if (f and islib(f)) or any(islib(f) for f in os.listdir(root)):
                logger.info('adding to LD_LIBRARY_PATH: %s',root)
                if 'LD_LIBRARY_PATH' in os.environ:
                    if root in os.environ['LD_LIBRARY_PATH'].split(':'):
                        return  # already present
                    os.environ['LD_LIBRARY_PATH'] = root+':'+os.environ['LD_LIBRARY_PATH']
                else:
                    os.environ['LD_LIBRARY_PATH'] = root
            else:
                logger.debug('no libs in %s',root)

        def addToPythonPath(root):
            if glob.glob(os.path.join(root,'*.py')):
                logger.info('adding to PYTHONPATH: %s',root)
                if 'PYTHONPATH' in os.environ:
                    if root in os.environ['PYTHONPATH'].split(':'):
                        return  # already present
                    os.environ['PYTHONPATH'] = root+':'+os.environ['PYTHONPATH']
                else:
                    os.environ['PYTHONPATH'] = root
            else:
                logger.debug('no python files: %s',root)
        if os.path.isdir(local):
            # build search list
            search_list = [local]
            search_list.extend(glob.glob(os.path.join(local,'lib*')))
            search_list.extend(glob.glob(os.path.join(local,'lib*/python*/*-packages')))
            if class_obj['libs'] is not None:
                search_list.extend(os.path.join(local,x) for x in class_obj['libs'].split(':'))
            for s in search_list:
                if not os.path.isdir(s):
                    continue
                addToPythonPath(s)
                ldpath(s)
        elif os.path.exists(local):
            root, f = os.path.split(local)
            if f.endswith('.py'):
                if root not in sys.path:
                    addToPythonPath(root)
            else:
                # check for binary library
                ldpath(root,f)
        # modify environment variables
        logger.info('env_vars = %s',class_obj['env_vars'])
        if class_obj['env_vars']:
            for e in class_obj['env_vars'].split(';'):
                try:
                    k,v = e.split('=')
                except ValueError as e:
                    logger.warning('bad env variable: %s',e)
                    continue
                v = v.replace('$CLASS',local)
                logger.info('setting envvar: %s = %s',k,v)
                if k in os.environ:
                    os.environ[k] = v+':'+os.environ[k]
                else:
                    os.environ[k] = v


# Run Functions #


async def runtask(cfg, globalenv, task, logger=None):
    """Run the specified task"""
    if not task:
        raise Exception('No task provided')
    if not logger:
        logger = logging

    # set up task_temp
    if not os.path.exists('task_temp'):
        os.mkdir('task_temp')
    globalenv['task_temp'] = os.path.join(os.getcwd(),'task_temp')

    # set up stats
    stats = {}

    # check if we have any files in the task_files API
    if task['task_files'] and ((not cfg.config['options']['offline']) or cfg.config['options']['offline_transfer']):
        files = await cfg.rpc.task_files(cfg.config['options']['dataset_id'],
                                         cfg.config['options']['task_id'])
        task['data'].extend(files)

    try:
        # set up local env
        async with SetupEnv(cfg, task, globalenv, logger=logger) as env:
            # run trays
            for tray in task['trays']:
                tmpstat = {}
                async for proc in runtray(cfg, env, tray, stats=tmpstat, logger=logger):
                    yield proc
                if len(tmpstat) > 1:
                    stats[tray['name']] = tmpstat
                elif len(tmpstat) == 1:
                    stats[tray['name']] = tmpstat[list(tmpstat.keys())[0]]
    finally:
        # destroy task temp
        try:
            functions.removedirs('task_temp')
        except Exception as e:
            logger.warning('error removing task_temp directory: %r',
                           e, exc_info=True)

    globalenv['stats']['tasks'].append(stats)


async def runtray(cfg, globalenv,tray,stats={}, logger=None):
    """Run the specified tray"""
    if not tray:
        raise Exception('No tray provided')
    if not logger:
        logger = logging

    # set up tray_temp
    if not os.path.exists('tray_temp'):
        os.mkdir('tray_temp')
    globalenv['tray_temp'] = os.path.join(os.getcwd(),'tray_temp')

    # run iterations
    try:
        tmpenv = globalenv.copy()
        for i in range(tray['iterations']):
            # set up local env
            cfg.config['options']['iter'] = i
            tmpstat = {}
            async with SetupEnv(cfg, tray, tmpenv, logger=logger) as env:
                # run modules
                for module in tray['modules']:
                    async for proc in runmodule(cfg, env, module, stats=tmpstat, logger=logger):
                        yield proc
            stats[i] = tmpstat

    finally:
        # destroy tray temp
        try:
            functions.removedirs('tray_temp')
        except Exception as e:
            logger.warning('error removing tray_temp directory: %s',
                           str(e), exc_info=True)


async def runmodule(cfg, globalenv, module, stats={}, logger=None):
    """Run the specified module"""
    if not module:
        raise Exception('No module provided')
    if not logger:
        logger = logging

    # set up local env
    module = module.copy()
    async with SetupEnv(cfg, module, globalenv, logger=logger) as env:
        if module['running_class']:
            module['running_class'] = cfg.parseValue(module['running_class'],env)
        if module['args']:
            module['args'] = cfg.parseObject(module['args'],env)
        if module['src']:
            module['src'] = cfg.parseValue(module['src'],env)
        if module['env_shell']:
            module['env_shell'] = cfg.parseValue(module['env_shell'],env)
        if module['configs']:
            # parse twice to make sure it's parsed, even if it starts as a string
            module['configs'] = cfg.parseObject(module['configs'],env)
            module['configs'] = cfg.parseObject(module['configs'],env)

        # make subprocess to run the module
        async with ForkModule(cfg, env, module, logger=logger, stats=stats) as process:
            # yield process back to pilot or driver, so it can be killed
            yield process


class ForkModule:
    """
    Modules are run in a forked process to prevent segfaults from killing IceProd.
    Their stdout and stderr is dumped into the log file with prefixes on each
    line to designate its source.  Any error or the return value is returned to
    the main process via a Queue.

    If a module defines a src, that is assumed to be a Class which should be
    added to the env.  The running_class is where the exact script or binary
    is chosen.  It can match several things:

    * A fully defined python module.class import (also takes module.function)
    * A python class defined in the src provided
    * A regular python script
    * An executable of some type (this is run in a subprocess with shell
      execution disabled)
    """
    def __init__(self, cfg, env, module, logger=None, stats=None):
        self.cfg = cfg
        self.env = env
        self.module = module
        if not logger:
            logger = logging
        self.logger = logger
        self.stats = stats if stats else {}
        self.proc = None

        self.error_filename = constants['task_exception']
        self.stats_filename = constants['stats']
        if 'subprocess_dir' in cfg.config['options'] and cfg.config['options']['subprocess_dir']:
            subdir = cfg.config['options']['subprocess_dir']
            self.error_filename = os.path.join(subdir, self.error_filename)
            self.stats_filename = os.path.join(subdir, self.stats_filename)

        if os.path.exists(self.error_filename):
            os.remove(self.error_filename)

    async def __aenter__(self):
        module_src = None
        if self.module['src']:
            if not functions.isurl(self.module['src']):
                module_src = self.module['src']
            else:
                # get script to run
                c = dataclasses.Class()
                c['src'] = self.module['src']
                c['name'] = os.path.basename(c['src'])
                if '?' in c['name']:
                    c['name'] = c['name'][:c['name'].find('?')]
                elif '#' in c['name']:
                    c['name'] = c['name'][:c['name'].find('#')]
                await setupClass(self.env,c,logger=self.logger)
                if c['name'] not in self.env['classes']:
                    raise Exception('Failed to install class %s'%c['name'])
                module_src = self.env['classes'][c['name']]

        # set up env_shell
        env_shell = None
        if self.module['env_shell']:
            env_shell = self.module['env_shell'].split()
            self.logger.info('searching for env_shell at %r', env_shell[0])
            if not os.path.exists(env_shell[0]):
                env_class = env_shell[0].split('/')[0]
                self.logger.info('searching for env_shell as %r class', env_class)
                if env_class in self.env['classes']:
                    env_tmp = env_shell[0].split('/')
                    env_tmp[0] = self.env['classes'][env_class]
                    env_shell[0] = '/'.join(env_tmp)
                else:
                    self.logger.info('attempting to download env_shell')
                    c = dataclasses.Class()
                    c['src'] = env_shell[0]
                    c['name'] = os.path.basename(c['src'])
                    await setupClass(self.env,c,logger=self.logger)
                    if c['name'] not in self.env['classes']:
                        raise Exception('Failed to install class %s'%c['name'])
                    env_shell[0] = self.env['classes'][c['name']]

        if module_src:
            self.logger.warning('running module \'%s\' with src %s',
                                self.module['name'], module_src)
        else:
            self.logger.warning('running module \'%s\' with class %s',
                                self.module['name'], self.module['running_class'])

        # set up the args
        args = self.module['args']
        if args:
            self.logger.warning('args=%s',args)
            if args and isinstance(args,dataclasses.String) and args[0] in ('{','['):
                args = json_decode(args)
            if args and isinstance(args, dict) and set(args) == {'args','kwargs'}:
                args = self.cfg.parseObject(args, self.env)
            elif isinstance(args,dataclasses.String):
                args = {"args":[self.cfg.parseValue(x,self.env) for x in args.split()],"kwargs":{}}
            elif isinstance(args,list):
                args = {"args":[self.cfg.parseValue(x,self.env) for x in args],"kwargs":{}}
            elif isinstance(args,dict):
                args = {"args":[],"kwargs":self.cfg.parseObject(args,self.env)}
            else:
                raise Exception('args is unknown type')

        # set up the environment
        cmd = []
        if env_shell:
            cmd.extend(env_shell)

        kwargs = {'close_fds': True}
        if 'subprocess_dir' in self.cfg.config['options'] and self.cfg.config['options']['subprocess_dir']:
            subdir = self.cfg.config['options']['subprocess_dir']
            if not os.path.exists(subdir):
                os.makedirs(subdir)
            kwargs['cwd'] = subdir
        else:
            kwargs['cwd'] = os.getcwd()
        self.stdout = open(os.path.join(kwargs['cwd'], constants['stdout']), 'ab')
        self.stderr = open(os.path.join(kwargs['cwd'], constants['stderr']), 'ab')
        kwargs['stdout'] = self.stdout
        kwargs['stderr'] = self.stderr

        # set up configs
        if self.module['configs']:
            for filename in self.module['configs']:
                self.logger.info('creating config %r', filename)
                with open(os.path.join(kwargs['cwd'], filename),'w') as f:
                    f.write(json_encode(self.module['configs'][filename]))

        # run the module
        if self.module['running_class']:
            self.logger.info('run as a class using the helper script')
            exe_helper = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                      'exe_helper.py')
            cmd.extend(['python', exe_helper, '--classname',
                        self.module['running_class']])
            if self.env['options']['debug']:
                cmd.append('--debug')
            if module_src:
                cmd.extend(['--filename', module_src])
            if args:
                args_filename = constants['args']
                if 'cwd' in kwargs:
                    args_filename = os.path.join(kwargs['cwd'], args_filename)
                with open(args_filename,'w') as f:
                    f.write(json_encode(args))
                cmd.append('--args')
        elif module_src:
            self.logger.info('run as a script directly')
            if args:
                def splitter(a,b):
                    ret = ('-%s' if len(str(a)) <= 1 else '--%s')%str(a)
                    if b is None:
                        return ret
                    else:
                        return ret+'='+str(b)
                args = args['args']+[splitter(a,args['kwargs'][a]) for a in args['kwargs']]

                # force args to string
                def toStr(a):
                    if isinstance(a,(bytes,str)):
                        return a
                    else:
                        return str(a)
                args = [toStr(a) for a in args]
            else:
                args = []

            shebang = False
            if os.path.exists(module_src):
                try:
                    with open(module_src) as f:
                        if f.read(10).startswith('#!'):
                            # shebang found
                            mode = os.stat(module_src).st_mode
                            if not (mode & stat.S_IXUSR):
                                os.chmod(module_src, mode | stat.S_IXUSR)
                            shebang = True
                except Exception:
                    self.logger.warning('cannot get shebang for %s', module_src,
                                        exc_info=True)

            if (not shebang) and module_src[-3:] == '.py':
                # call as python script
                cmd.extend(['python', module_src]+args)
            elif (not shebang) and module_src[-3:] == '.sh':
                # call as shell script
                cmd.extend(['/bin/sh', module_src]+args)
            else:
                # call as regular executable
                cmd.extend([module_src]+args)
        else:
            self.logger.error('module is missing class and src')
            raise Exception('error running module')

        self.logger.warning('subprocess cmd=%r',cmd)
        if self.module['env_clear']:
            # must be on cvmfs-like environ for this to apply
            env = {'PYTHONNOUSERSITE':'1'}
            if 'SROOT' in os.environ:
                prefix = os.environ['SROOT']
            elif 'ICEPRODROOT' in os.environ:
                prefix = os.environ['ICEPRODROOT']
            else:
                prefix = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
            for k in os.environ:
                if k in ('OPENCL_VENDOR_PATH','http_proxy','TMP','TMPDIR','_CONDOR_SCRATCH_DIR'):
                    # pass through unchanged
                    env[k] = os.environ[k]
                elif ('sroot' in k.lower() or 'iceprod' in k.lower() or
                      k in ('CUDA_VISIBLE_DEVICES','COMPUTE','GPU_DEVICE_ORDINAL')):
                    # don't pass these at all
                    pass
                else:
                    # filter SROOT out of environ
                    ret = [x for x in os.environ[k].split(':') if x.strip() and (not x.startswith(prefix)) and 'iceprod' not in x.lower()]
                    if ret:
                        env[k] = ':'.join(ret)
            # handle resource environment
            if 'resources' in self.cfg.config['options']:
                Resources.set_env(self.cfg.config['options']['resources'], env)
            self.logger.warning('env = %r', env)
            kwargs['env'] = env

        self.proc = await asyncio.create_subprocess_exec(*cmd, **kwargs)
        return self.proc

    async def __aexit__(self, exc_type, exc, tb):
        try:
            self.stdout.close()
            self.stderr.close()
        except Exception:
            pass
        if not exc_type:
            # now clean up after process
            if self.proc and self.proc.returncode:
                self.logger.warning('return code: {}'.format(self.proc.returncode))
                try:
                    with open(self.error_filename, 'rb') as f:
                        e = pickle.load(f)
                except Exception:
                    self.logger.warning('cannot load exception info from failed module', )
                    raise Exception('module failed')
                else:
                    if isinstance(e, Exception):
                        raise e
                    else:
                        raise Exception(str(e))

            # get stats, if available
            if os.path.exists(self.stats_filename):
                try:
                    new_stats = pickle.load(open(self.stats_filename, 'rb'))
                    if self.module['name']:
                        self.stats[self.module['name']] = new_stats
                    else:
                        self.stats.update(new_stats)
                except Exception:
                    self.logger.warning('cannot load stats info from module')



"""
Interface for configuring modules
"""

from __future__ import absolute_import, division, print_function

import logging

from statsd import TCPStatsClient
import requests

try:
    import boto3
    import botocore.client
except ImportError:
    boto3 = None

from rest_tools.client import ClientCredentialsAuth

logger = logging.getLogger('module')


class FakeStatsClient(object):
    def __init__(self, *args, **kwargs):
        self._prefix = ''

    def __getattr__(self, name):
        def foo(*args, **kwargs):
            pass
        return foo


class StatsClientIgnoreErrors(object):
    def __init__(self, *args, **kwargs):
        kwargs['timeout'] = 0.1
        self._statsclient = TCPStatsClient(*args, **kwargs)

    def __getattr__(self, name):
        def foo(*args, **kwargs):
            try:
                try:
                    return getattr(self._statsclient, name)(*args, **kwargs)
                except BrokenPipeError:
                    self._statsclient.reconnect()
                    return getattr(self._statsclient, name)(*args, **kwargs)
            except Exception:
                logging.info('StatsClient dropped %s %r %r', name, args, kwargs,
                             exc_info=True)
        return foo


class ElasticClient(object):
    def __init__(self, hostname, basename='iceprod'):
        self.session = requests.Session()
        # handle auth
        if '@' in hostname:
            self.session.auth = tuple(hostname.split('@')[0].split('://')[1].split(':'))
            hostname = hostname.split('://',1)[0]+'://'+hostname.split('@',1)[1]
        # try a connection
        r = self.session.get(hostname, timeout=5)
        r.raise_for_status()
        # concat hostname and basename
        self.hostname = hostname+'/'+basename+'_'

    def head(self, name, index_name):
        try:
            r = self.session.head(self.hostname+name+'/item/'+index_name, timeout=5)
            r.raise_for_status()
        except Exception:
            return False
        else:
            return True

    def get(self, name, index_name):
        r = self.session.get(self.hostname+name+'/item/'+index_name, timeout=5)
        r.raise_for_status()
        return r.json()

    def post(self, name, index_name, data):
        r = self.session.post(self.hostname+name+'/item/'+index_name, timeout=5)
        r.raise_for_status()
        return r.json()

    def put(self, name, index_name, data):
        r = None
        try:
            kwargs = {'timeout':5}
            if isinstance(data,dict):
                kwargs['json'] = data
            else:
                kwargs['data'] = data
            r = self.session.put(self.hostname+name+'/item/'+index_name, **kwargs)
            r.raise_for_status()
        except Exception:
            logger.warning('cannot put to elasticsearch: %s%s/%s',
                           self.hostname, name, index_name, exc_info=True)
            if r:
                logger.info('%r',r.content)


class module(object):
    """
    This is an abstract class representing a server module.

    :param cfg: An :class:`IceProdConfig`.
    :param executor: A :class:`concurrent.futures.ThreadPoolExecutor`.
    :param modules: A dict of other module's public services.
    """
    def __init__(self, cfg, executor, modules):
        self.cfg = cfg
        self.executor = executor
        self.statsd = FakeStatsClient()
        self.elasticsearch = FakeStatsClient()
        self.s3 = None
        self.rest_client = None
        self.cred_client = None
        self.modules = modules
        self.service = {'start': self.start,
                        'stop': self.stop,
                        'kill': self.kill}

    def start(self):
        """
        Set up a module.
        """
        logger.warning('starting module %s', self.__class__.__name__)
        if 'statsd' in self.cfg and self.cfg['statsd']:
            try:
                addr = self.cfg['statsd']
                port = 8125
                if ':' in addr:
                    addr,port = addr.split(':')
                    port = int(port)
                self.statsd = StatsClientIgnoreErrors(
                    addr,
                    port=port,
                    prefix=self.cfg['site_id']+'.'+self.__class__.__name__
                )
            except Exception:
                logger.warning('failed to connect to statsd: %r',
                               self.cfg['statsd'], exc_info=True)

        if 'elasticsearch' in self.cfg and self.cfg['elasticsearch']:
            try:
                self.elasticsearch = ElasticClient(self.cfg['elasticsearch'])
            except Exception:
                logger.warning('failed to connect to elasicsearch: %r',
                               self.cfg['elasticsearch'], exc_info=True)

        if (boto3 and 's3' in self.cfg and 'access_key' in self.cfg['s3'] and
                'secret_key' in self.cfg['s3']):
            try:
                self.s3 = boto3.client(
                    's3',
                    'us-east-1',
                    aws_access_key_id=self.cfg['s3']['access_key'],
                    aws_secret_access_key=self.cfg['s3']['secret_key'],
                    config=botocore.client.Config(max_pool_connections=101)
                )
            except Exception:
                logger.warning('failed to connect to s3: %r',
                               self.cfg['s3'], exc_info=True)
        if ('rest_api' in self.cfg and 'url' in self.cfg['rest_api']
                and 'oauth_url' in self.cfg['rest_api']
                and 'oauth_client_id' in self.cfg['rest_api']
                and 'oauth_client_secret' in self.cfg['rest_api']):
            try:
                self.rest_client = ClientCredentialsAuth(
                    address=self.cfg['rest_api']['url'],
                    token_url=self.cfg['rest_api']['oauth_url'],
                    client_id=self.cfg['rest_api']['oauth_client_id'],
                    client_secret=self.cfg['rest_api']['oauth_client_secret'],
                )
                self.cred_client = ClientCredentialsAuth(
                    address=self.cfg['rest_api']['cred_url'],
                    token_url=self.cfg['rest_api']['oauth_url'],
                    client_id=self.cfg['rest_api']['oauth_client_id'],
                    client_secret=self.cfg['rest_api']['oauth_client_secret'],
                )
            except Exception:
                logger.warning('failed to connect to rest api: %r',
                               self.cfg['rest_api'].get('url',''), exc_info=True)

    def stop(self):
        logger.warning('stopping module %s', self.__class__.__name__)

    def kill(self):
        logger.warning('killing module %s', self.__class__.__name__)



"""
Some general functions used by the iceprod server
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import logging
from pkgutil import get_loader
import importlib
import subprocess


def find_module_recursive(name, path=None):
    """ Recursively search for submodule. Submodules must be separated with a '.' """
    import imp
    res = None
    for x in name.split('.'):
        res = imp.find_module(x, path)
        path = [res[1]]
    return res


def listmodules(package_name=''):
    """List modules in a package or directory"""
    file, pathname, description = find_module_recursive(package_name)
    if file:
        # Not a package
        return []
    ret = []
    for module in os.listdir(pathname):
        if module.endswith('.py') and module != '__init__.py':
            tmp = os.path.splitext(module)[0]
            ret.append(package_name+'.'+tmp)
    return ret


def run_module(name,*args,**kwargs):
    """Import and start the module"""
    class_name = name.rsplit('.',1)[1]
    x = importlib.import_module(name)
    return (getattr(x,class_name))(*args,**kwargs)


class GlobalID(object):
    """Global ID configuration and generation"""
    import string
    # never change these settings, otherwise all old ids will fail
    CHARS = string.ascii_letters+string.digits
    CHARS_LEN = len(CHARS)
    # define dict to make reverse lookup super fast
    INTS_DICT = {c:i for i,c in enumerate(CHARS)}
    IDLEN = 15
    MAXSITEID = 10**10
    MAXLOCALID = 10**15

    @classmethod
    def int2char(cls,i):
        if not isinstance(i,int) or i < 0:  # only deal with positive ints
            logging.warning('bad input to int2char: %r',i)
            raise Exception('bad input to int2char')
        out = ''
        while i >= 0:
            out += cls.CHARS[i%cls.CHARS_LEN]
            i = i//cls.CHARS_LEN - 1
        return out[::-1]

    @classmethod
    def char2int(cls,c):
        if not isinstance(c,str) or len(c) < 1:  # only deal with string
            logging.warning('bad input to char2int: %r',c)
            raise Exception('bad input to char2int')
        out = -1
        for i,cc in enumerate(reversed(c)):
            if cc not in cls.CHARS:
                raise Exception('non-char input to chars2int')
            out += (cls.INTS_DICT[cc]+1)*(cls.CHARS_LEN**i)
        return out

    @classmethod
    def siteID_gen(cls):
        """Generate a new site id"""
        import random
        return cls.int2char(random.randint(0,cls.MAXSITEID-1))

    @classmethod
    def globalID_gen(cls,id,site_id):
        """Generate a new global id given a local id and site id"""
        if isinstance(id,str):
            id = cls.char2int(id)
        elif not isinstance(id,int):
            raise Exception('id is not a string, int, or long')
        if isinstance(site_id,str):
            return cls.int2char(cls.char2int(site_id)*cls.MAXLOCALID+id)
        elif isinstance(site_id,int):
            return cls.int2char(site_id*cls.MAXLOCALID+id)
        else:
            raise Exception('Site id is not a string, int, or long')

    @classmethod
    def localID_ret(cls,id,type='str'):
        """Retrieve a local id from a global id"""
        ret = cls.char2int(id) % cls.MAXLOCALID
        if type == 'str':
            ret = cls.int2char(ret)
        return ret

    @classmethod
    def siteID_ret(cls,id,type='str'):
        """Retrieve a site id from a global id"""
        ret = cls.char2int(id) // cls.MAXLOCALID
        if type == 'str':
            ret = cls.int2char(ret)
        return ret


def salt(length=2):
    """Returns a string of random letters"""
    import string
    import random
    letters = string.ascii_letters+string.digits
    return ''.join([random.SystemRandom().choice(letters) for _ in range(length)])


class KwargConfig(object):
    """A way to validate kwargs passed in to a class"""
    def __init__(self):
        # defaults
        self._cfg = {}
        self._cfg_types = {}

    def validate(self,kwargs):
        # setup cfg variables
        for s in kwargs.keys():
            v = kwargs[s]
            if not isinstance(s,str):
                raise Exception('parameter name %s is not a string'%(str(s)))
            if s not in self._cfg:
                logging.warning('%s is not a valid arg',s)
                continue
            t = self._cfg_types[s]
            if t in ('str','file','dir'):
                if not isinstance(v,str):
                    raise Exception('%s is not a string'%(str(s)))
                if t in ('file','dir'):
                    v = os.path.expanduser(os.path.expandvars(v))
                    if t == 'file' and not ('_file' in s or '_log' in s):
                        try:
                            os.path.exists(v)
                        except Exception:
                            raise Exception('parameter %s with filepath %s does not exist'%(s,v))
            elif t == 'int':
                if not isinstance(v,int):
                    raise Exception('%s is not an int'%(str(s)))
            elif t == 'float':
                if not isinstance(v,float):
                    raise Exception('%s is not a float'%(str(s)))
            else:
                raise Exception('%s has an unknown type'%(str(s)))
            self._cfg[s] = v

        # make directories
        for c in self._cfg_types:
            if self._cfg_types[c] == 'file':
                d = os.path.dirname(self._cfg[c])
                if not os.path.isdir(d):
                    os.makedirs(d)
            if self._cfg_types[c] == 'dir':
                d = self._cfg[c]
                if not os.path.isdir(d):
                    os.makedirs(d)


def get_pkgdata_filename(package, resource):
    """Get a filename for a resource bundled within the package"""
    loader = get_loader(package)
    if loader is None or not hasattr(loader, 'get_data'):
        return None
    mod = sys.modules.get(package) or loader.load_module(package)
    if mod is None or not hasattr(mod, '__file__'):
        return None

    # Modify the resource name to be compatible with the loader.get_data
    # signature - an os.path format "filename" starting with the dirname of
    # the package's __file__
    parts = resource.split('/')
    parts.insert(0, os.path.dirname(mod.__file__))
    return os.path.join(*parts)


def get_pkg_binary(package, binary):
    """Try finding the binary path based on the python package"""
    try:
        loader = get_loader(package)
        f = loader.get_filename()
        while f and 'lib' in f:
            f = os.path.dirname(f)
        filepath = os.path.join(f,'bin',binary)
        if os.path.exists(filepath):
            return filepath
        filepath = os.path.join(f,'sbin',binary)
        if os.path.exists(filepath):
            return filepath
    except Exception:
        pass

    # try going up from sys.argv[0]
    try:
        f = os.path.abspath(sys.argv[0])
        while f and 'iceprod' in f:
            filepath = os.path.join(f,'bin',binary)
            if os.path.exists(filepath):
                return filepath
            f = os.path.dirname(f)
        filepath = os.path.join(f,'bin',binary)
        if os.path.exists(filepath):
            return filepath
    except Exception:
        pass

    # try just asking the shell
    try:
        filepath = subprocess.check_output(["which",binary]).decode('utf-8').strip('\n')
        if os.path.exists(filepath):
            return filepath
    except Exception:
        pass
    return None



"""
Help run class-based modules, including iceprod modules.

This is run in a subprocess to help set up the environment,
as well as contain any crashes.

Note that this file should be backward-compatible with python 2.7+,
as it will be run under the user's environment.
"""

from __future__ import absolute_import, division, print_function

import os
import imp
import inspect
import logging
import importlib
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

try:
    import cPickle as pickle
except Exception:
    import pickle

from json import loads as json_decode

# from iceprod.core import constants
constants_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),'__init__.py')
try:
    spec = importlib.util.spec_from_file_location('constants', constants_path)
    constants_mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(constants_mod)
except AttributeError:
    constants_mod = imp.load_source('constants',constants_path)
constants = constants_mod.constants

try:
    String = basestring
except NameError:
    String = str


def get_args():
    """Read json of [args, kwargs] from the std args file"""
    with open(constants['args']) as f:
        data = f.read()
        logging.debug('get_args raw: %r',data)
        return json_decode(data)


def unicode_to_ascii(obj):
    if isinstance(obj,String):
        return str(obj)
    elif isinstance(obj,dict):
        ret = {}
        for k in obj:
            ret[unicode_to_ascii(k)] = unicode_to_ascii(obj[k])
        return ret
    elif isinstance(obj,set):
        return set(unicode_to_ascii(k) for k in obj)
    elif isinstance(obj,Iterable):
        return [unicode_to_ascii(k) for k in obj]
    else:
        return obj


def run(classname, filename=None, args=False, debug=False):
    logging.basicConfig(level=logging.DEBUG if debug else logging.WARN)
    logging.warning('exe_helper(%s)', classname)

    if not classname:
        raise Exception('classname is missing')
    class_args = {'args':[],'kwargs':{}}
    if args:
        class_args = get_args()
        logging.info('args: %r', class_args)
    class_args = unicode_to_ascii(class_args)
    parts = classname.rsplit('.',1)
    if len(parts) == 1:
        p,cl = os.path.basename(filename),parts[0]
    else:
        p,cl = parts
    if filename:
        logging.info('try loading from source: %s', filename)
        mod = imp.load_source(p, filename)
        class_obj = getattr(mod,cl)
    else:
        logging.info('try regular import: %s.%s', p, cl)
        mod = __import__(p,globals(),locals(),[cl])
        class_obj = getattr(mod,cl)

    if (inspect.isclass(class_obj) and
            any(True for c in inspect.getmro(class_obj) if c.__name__ == 'IPBaseClass')):
        logging.info('IceProd v1 class')
        instance = class_obj()
        for k in class_args['kwargs']:
            instance.SetParameter(k,class_args['kwargs'][k])
        stats = {}
        ret = instance.Execute(stats)
        if stats:
            pickle.dump(stats,open(constants['stats'],'wb'))
        if ret:
            raise Exception('Execute() returned %r'%ret)
    else:
        logging.info('regular callable')
        class_obj(*class_args['args'],**class_args['kwargs'])


def main():
    import argparse
    parser = argparse.ArgumentParser(description='IceProd Core')
    parser.add_argument('-c','--classname', type=str, default=None,
                        help='Specify class name')
    parser.add_argument('-f','--filename', type=str, default=None,
                        help='Specify file to find the class in')
    parser.add_argument('-a','--args', action='store_true', default=False,
                        help='Enable arg file detection')
    parser.add_argument('-d','--debug', action='store_true', default=False,
                        help='Enable debug actions and logging')
    args = vars(parser.parse_args())
    try:
        run(**args)
    except Exception as e:
        with open(constants['task_exception'],'wb') as f:
            pickle.dump(e,f)
        raise


if __name__ == '__main__':
    main()



"""
Reset tasks that are not active (in a pilot).

Check all the tasks that are processing, and compare with the
tasks in pilots.  Reset the difference.
"""

import argparse
import asyncio
from datetime import datetime
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.util import str2datetime

logger = logging.getLogger('non_active_tasks')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    try:
        datasets = await rest_client.request('GET', '/dataset_summaries/status')
        dataset_ids = []
        if 'processing' in datasets:
            dataset_ids.extend(datasets['processing'])
        if 'truncated' in datasets:
            dataset_ids.extend(datasets['truncated'])
        pilots = await rest_client.request('GET', '/pilots', {'keys': 'pilot_id|tasks'})
        task_ids_in_pilots = set()
        for p in pilots.values():
            if 'tasks' in p and p['tasks']:
                task_ids_in_pilots.update(p['tasks'])
        dataset_tasks = {}
        for dataset_id in dataset_ids:
            dataset_tasks[dataset_id] = await rest_client.request('GET', '/datasets/{}/task_summaries/status'.format(dataset_id))

        async def reset(dataset_id, task_id):
            args = {'status':'reset'}
            await rest_client.request('PUT', f'/datasets/{dataset_id}/tasks/{task_id}/status', args)
            data = {
                'name': 'stdlog',
                'task_id': task_id,
                'dataset_id': dataset_id,
                'data': 'task status = processing, but not found in any pilot',
            }
            await rest_client.request('POST', '/logs', data)
            data.update({'name':'stdout', 'data': ''})
            await rest_client.request('POST', '/logs', data)
            data.update({'name':'stderr', 'data': ''})
            await rest_client.request('POST', '/logs', data)

        async def update_pilot(pilot_id, tasks):
            args = {'tasks': tasks}
            await rest_client.request('PATCH', '/pilots/{}'.format(pilot_id), args)

        async def delete_pilot(pilot_id):
            await rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))

        awaitables = set()
        reset_pilots = set()
        for dataset_id in dataset_ids:
            tasks = dataset_tasks[dataset_id]
            if 'processing' in tasks:
                reset_tasks = set(tasks['processing'])-task_ids_in_pilots
                for task_id in reset_tasks:
                    args = {'keys': 'status|status_changed'}
                    task = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}', args)
                    # check status, and that we haven't just changed status
                    if task['status'] == 'processing' and (datetime.utcnow()-str2datetime(task['status_changed'])).total_seconds() > 600:
                        logger.info('dataset %s reset task %s', dataset_id, task_id)
                        awaitables.add(reset(dataset_id,task_id))

            for k in tasks.keys():
                if k in ('reset', 'waiting', 'failed', 'suspended'):
                    for task_id in task_ids_in_pilots.intersection(tasks[k]):
                        args = {'keys': 'status|status_changed'}
                        task = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}', args)
                        # check status, and that we haven't just changed status
                        if task['status'] in ('reset', 'waiting', 'failed', 'suspended') and (datetime.utcnow()-str2datetime(task['status_changed'])).total_seconds() > 600:
                            reset_pilots.add(task_id)

        for p in pilots.values():
            if 'tasks' in p:
                updated_tasks = set(p['tasks']) - reset_pilots
                if not updated_tasks:
                    awaitables.add(delete_pilot(p['pilot_id']))
                elif updated_tasks != set(p['tasks']):
                    awaitables.add(delete_pilot(p['pilot_id'], updated_tasks))

        for fut in asyncio.as_completed(awaitables):
            try:
                await fut
            except Exception:
                logger.error('error resetting non-active tasks', exc_info=True)
                if debug:
                    raise
    except Exception:
        logger.error('error resetting non-active tasks', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""Pilot functionality"""

from __future__ import absolute_import, division, print_function

import os
import sys
import math
import time
import logging
import tempfile
import shutil
import random
from datetime import datetime
from glob import glob
import signal
import traceback
import asyncio
import concurrent.futures

import iceprod
from iceprod.core.functions import gethostname
from iceprod.core import constants
from iceprod.core.resources import Resources
import iceprod.core.logger

logger = logging.getLogger('pilot')

try:
    import psutil
except ImportError:
    psutil = None

try:
    from setproctitle import setproctitle
except ImportError:
    def setproctitle(name):
        pass


class Pilot:
    """
    A pilot task runner.

    The pilot allows multiple tasks to run in sequence or parallel.
    It keeps track of resource usage, killing anything that goes over
    requested amounts.

    Use as an async context manager::

        async with Pilot(*args) as p:
            await p.run()

    Args:
        config (dict): the configuration dictionary
        runner (callable): the task/config runner
        pilot_id (str): the pilot id
        rpc (:py:class:`iceprod.core.exe_json.ServerComms`): RPC to server
        debug (bool): debug mode (default False)
        run_timeout (int): how often to check if a task is running
        backoff_delay (int): what constant delay to use for backoff
        backoff_factor (int): what starting delay to use for exponential backoff
        download_delay (int): min delay between each task download attempt
        resource_interval (float): seconds between resouce measurements
        restrict_site (bool): restrict running tasks to explicitly requiring this site
    """
    def __init__(self, config, runner, pilot_id, rpc=None, debug=False,
                 run_timeout=180, backoff_delay=60, backoff_factor=1,
                 download_delay=60, resource_interval=1.0, restrict_site=False):
        self.config = config
        self.runner = runner
        self.pilot_id = pilot_id
        self.hostname = gethostname()
        self.rpc = rpc
        self.debug = debug
        self.run_timeout = run_timeout
        self.backoff_delay = backoff_delay
        self.backoff_factor = backoff_factor
        self.download_delay = download_delay
        self.resource_interval = resource_interval
        self.query_params = {}
        self.last_download = None

        self.running = True
        self.tasks = {}

        try:
            setproctitle('iceprod2_pilot({})'.format(pilot_id))
        except Exception:
            pass

        logger.warning('pilot_id: %s', self.pilot_id)
        logger.warning('hostname: %s', self.hostname)

        # hint at resources for pilot
        # don't pass them as raw, because that overrides condor
        if 'resources' in config['options']:
            for k in config['options']['resources']:
                v = config['options']['resources'][k]
                name = 'NUM_'+k.upper()
                if k in ('cpu','gpu'):
                    name += 'S'
                os.environ[name] = str(v)
        self.resources = Resources(debug=self.debug)
        if restrict_site:
            if not self.resources.site:
                logger.error('cannot restrict site, as the site is unknown')
            else:
                self.query_params['requirements.site'] = self.resources.site

        self.start_time = time.time()

    async def __aenter__(self):
        # update pilot status
        await self.rpc.update_pilot(
            self.pilot_id,
            tasks=[],
            host=self.hostname,
            version=iceprod.__version__,
            site=self.resources.site,
            start_date=datetime.utcnow().isoformat(),
            resources_available=self.resources.get_available(),
            resources_claimed=self.resources.get_claimed()
        )

        loop = asyncio.get_event_loop()
        # set up resource monitor
        if psutil:
            loop.create_task(self.resource_monitor())
        else:
            logger.warning('no psutil. not checking resource usage')

        # set up signal handler
        def handler(signum, frame):
            logger.critical('termination signal received')
            self.running = False
            self.term_handler()
        self.prev_signal = signal.signal(signal.SIGTERM, handler)

        return self

    async def __aexit__(self, exc_type, exc, tb):
        try:
            # make sure any child processes are dead
            self.hard_kill()

            if self.debug:
                # append out, err, log
                for dirs in glob('tmp*'):
                    for filename in (constants['stdout'], constants['stderr'],
                                     constants['stdlog']):
                        if os.path.exists(os.path.join(dirs,filename)):
                            with open(filename,'a') as f:
                                print('', file=f)
                                print('----',dirs,'----', file=f)
                                with open(os.path.join(dirs,filename)) as f2:
                                    print(f2.read(), file=f)

            await self.rpc.delete_pilot(self.pilot_id)
        except Exception:
            logger.error('error in aexit', exc_info=True)

        # restore previous signal handler
        signal.signal(signal.SIGTERM, self.prev_signal)

    def term_handler(self):
        """Handle a SIGTERM gracefully"""
        logger.info('checking resources after SIGTERM')
        overages = self.resources.check_claims()
        for task_id in list(self.tasks):
            task = self.tasks[task_id]
            try:
                if task_id in overages:
                    reason = overages[task_id]
                else:
                    reason = 'pilot SIGTERM'

                # clean up task
                used_resources = self.resources.get_final(task_id)
                self.clean_task(task_id)
                message = reason
                message += '\n\npilot SIGTERM\npilot_id: {}'.format(self.pilot_id)
                message += '\nhostname: {}'.format(self.hostname)
                kwargs = {
                    'resources': used_resources,
                    'reason': reason,
                    'message': message,
                }
                if 'dataset_id' in task['config']['options']:
                    kwargs['dataset_id'] = task['config']['options']['dataset_id']
                self.rpc.task_kill_sync(task_id, **kwargs)
            except Exception:
                pass

        # stop the pilot
        try:
            self.rpc.delete_pilot_sync(self.pilot_id)
        except Exception:
            pass
        sys.exit(1)

    def hard_kill(self):
        """Forcefully kill any child processes"""
        if psutil:
            # kill children correctly
            processes = psutil.Process().children(recursive=True)
            processes.reverse()
            for p in processes:
                try:
                    p.kill()
                except psutil.NoSuchProcess:
                    pass
                except Exception:
                    logger.warning('error killing process',
                                   exc_info=True)
        for task in self.tasks.values():
            try:
                task['p'].kill()
            except ProcessLookupError:
                logger.warning('error killing process',
                               exc_info=True)

    async def resource_monitor(self):
        """Monitor the tasks, killing any that go over resource limits"""
        try:
            sleep_time = self.resource_interval  # check every X seconds
            while self.running or self.tasks:
                logger.debug('pilot monitor - checking resource usage')
                start_time = time.time()

                try:
                    overages = self.resources.check_claims()
                    for task_id in overages:
                        used_resources = self.resources.get_peak(task_id)
                        logger.warning('kill %r for going over resources: %r',
                                       task_id, used_resources)
                        message = overages[task_id]
                        message += '\n\npilot_id: {}'.format(self.pilot_id)
                        message += '\nhostname: {}'.format(self.hostname)
                        kwargs = {
                            'resources': used_resources,
                            'reason': overages[task_id],
                            'message': message,
                        }
                        if 'dataset_id' in self.tasks[task_id]['config']['options']:
                            kwargs['dataset_id'] = self.tasks[task_id]['config']['options']['dataset_id']

                        self.clean_task(task_id)
                        await self.rpc.task_kill(task_id, **kwargs)
                except Exception:
                    logger.error('error in resource_monitor', exc_info=True)

                duration = time.time()-start_time
                logger.debug('sleep_time %.2f, duration %.2f',sleep_time,duration)
                if duration < sleep_time:
                    await asyncio.sleep(sleep_time-duration)
        except Exception:
            logger.error('pilot monitor died', exc_info=True)
        logger.warning('pilot monitor exiting')

    async def run(self):
        """Run the pilot"""
        download_errors = max_download_errors = 5
        iceprod_errors = 10
        task_errors = max_task_errors = int(10**math.log10(10+self.resources.total['cpu']))
        logger.info('max_errors: %d, %d', max_download_errors, max_task_errors)
        tasks_running = 0

        async def backoff():
            """Backoff for rate limiting"""
            delay = self.backoff_delay+self.backoff_factor*(1+random.random())
            logger.info('backoff %d', delay)
            await asyncio.sleep(delay)
            self.backoff_factor *= 2
        while self.running or self.tasks:
            while self.running:
                # retrieve new task(s)
                if self.last_download and time.time()-self.last_download < self.download_delay:
                    logger.warning('last download attempt too recent, backing off')
                    await asyncio.sleep(time.time()-self.last_download+self.download_delay)
                    break
                self.last_download = time.time()
                # if self.resources.total['gpu'] and not self.resources.available['gpu']:
                #     logger.info('gpu pilot with no gpus left - not queueing')
                #     break
                try:
                    task_configs = await self.rpc.download_task(
                        self.config['options']['gridspec'],
                        resources=self.resources.get_available(),
                        site=self.resources.site,
                        query_params=self.query_params
                    )
                except Exception:
                    download_errors -= 1
                    if download_errors < 1:
                        self.running = False
                        logger.warning('errors over limit, draining')
                    logger.error('cannot download task. current error count is %d',
                                 max_download_errors-download_errors, exc_info=True)
                    await backoff()
                    continue
                logger.info('task configs: %r', task_configs)

                if not task_configs:
                    logger.info('no task available')
                    if not self.tasks:
                        self.running = False
                        logger.warning('no task available, draining')
                    break
                else:
                    # start up new task(s)
                    for task_config in task_configs:
                        try:
                            task_id = task_config['options']['task_id']
                        except Exception:
                            iceprod_errors -= 1
                            if iceprod_errors < 1:
                                self.running = False
                                logger.warning('errors over limit, draining')
                            logger.error('error getting task_id from config')
                            break
                        try:
                            if 'resources' not in task_config['options']:
                                task_config['options']['resources'] = None
                            task_resources = self.resources.claim(task_id, task_config['options']['resources'])
                            task_config['options']['resources'] = task_resources
                        except Exception:
                            iceprod_errors -= 1
                            if iceprod_errors < 1:
                                self.running = False
                                logger.warning('errors over limit, draining')
                            logger.warning('error claiming resources %s', task_id,
                                           exc_info=True)
                            message = 'pilot_id: {}\nhostname: {}\n\n'.format(self.pilot_id, self.hostname)
                            message += traceback.format_exc()
                            kwargs = {
                                'reason': 'failed to claim resources',
                                'message': message,
                            }
                            if 'dataset_id' in task_config['options']:
                                kwargs['dataset_id'] = task_config['options']['dataset_id']
                            await self.rpc.task_kill(task_id, **kwargs)
                            break
                        try:
                            f = self.create_task(task_config)
                            task = await f.__anext__()
                            task['iter'] = f
                            self.tasks[task_id] = task
                        except Exception:
                            iceprod_errors -= 1
                            if iceprod_errors < 1:
                                self.running = False
                                logger.warning('errors over limit, draining')
                            logger.warning('error creating task %s', task_id,
                                           exc_info=True)
                            message = 'pilot_id: {}\nhostname: {}\n\n'.format(self.pilot_id, self.hostname)
                            message += traceback.format_exc()
                            kwargs = {
                                'reason': 'failed to create task',
                                'message': message,
                            }
                            if 'dataset_id' in task_config['options']:
                                kwargs['dataset_id'] = task_config['options']['dataset_id']
                            await self.rpc.task_kill(task_id, **kwargs)
                            self.clean_task(task_id)
                            break

                    # update pilot status
                    await self.rpc.update_pilot(self.pilot_id, tasks=list(self.tasks),
                                                resources_available=self.resources.get_available(),
                                                resources_claimed=self.resources.get_claimed())

                if (self.resources.available['cpu'] < 1
                        or self.resources.available['memory'] < 1
                        or (self.resources.total['gpu'] and not self.resources.available['gpu'])):
                    logger.info('no resources left, so wait for tasks to finish')
                    break
                # otherwise, backoff
                await backoff()

            # wait until we can queue more tasks
            while self.running or self.tasks:
                logger.info('wait while tasks are running. timeout=%r',self.run_timeout)
                start_time = time.time()
                while self.tasks and time.time()-self.run_timeout < start_time:
                    done,pending = await asyncio.wait([asyncio.create_task(task['p'].wait()) for task in self.tasks.values()],
                                                      timeout=self.resource_interval,
                                                      return_when=concurrent.futures.FIRST_COMPLETED)
                    if done:
                        break

                tasks_running = len(self.tasks)
                for task_id in list(self.tasks):
                    # check if any processes have died
                    proc = self.tasks[task_id]['p']
                    clean = False
                    if proc.returncode is not None:
                        f = self.tasks[task_id]['iter']
                        try:
                            task = await f.__anext__()
                        except StopAsyncIteration:
                            logger.warning('task %s finished', task_id)
                        except Exception:
                            logger.warning('task %s failed', task_id,
                                           exc_info=True)
                            task_errors -= 1
                        else:
                            logger.warning('task %s yielded again', task_id)
                            task['iter'] = f
                            self.tasks[task_id] = task
                            continue

                        # if we got here, the task is done
                        clean = True

                        # make sure the task is not running anymore
                        try:
                            await self.rpc.still_running(task_id)
                        except Exception:
                            pass
                        else:
                            logger.warning('task %s is still running, so killing it', task_id)
                            kwargs = {
                                'reason': 'task exited with return code {}'.format(proc.returncode),
                                'message': 'task exited with return code {}'.format(proc.returncode),
                                'resources': self.resources.get_final(task_id),
                            }
                            if 'dataset_id' in self.tasks[task_id]['config']['options']:
                                kwargs['dataset_id'] = self.tasks[task_id]['config']['options']['dataset_id']
                            await self.rpc.task_kill(task_id, **kwargs)
                    else:
                        # check if the DB has killed a task
                        try:
                            await self.rpc.still_running(task_id)
                        except Exception:
                            logger.warning('task %s killed by db', task_id)
                            kwargs = {
                                'reason': 'server kill',
                                'message': 'The server has marked the task as no longer running',
                            }
                            if 'dataset_id' in self.tasks[task_id]['config']['options']:
                                kwargs['dataset_id'] = self.tasks[task_id]['config']['options']['dataset_id']
                            await self.rpc.task_kill(task_id, **kwargs)
                            clean = True
                    if clean:
                        self.clean_task(task_id)
                if task_errors < 1:
                    self.running = False
                    logger.warning('errors over limit, draining')

                # update pilot status
                if (not self.tasks) or len(self.tasks) < tasks_running:
                    logger.info('%d tasks removed', tasks_running-len(self.tasks))
                    tasks_running = len(self.tasks)
                    await self.rpc.update_pilot(self.pilot_id, tasks=list(self.tasks),
                                                resources_available=self.resources.get_available(),
                                                resources_claimed=self.resources.get_claimed())
                    if self.running:
                        break
                elif (self.running and self.resources.available['cpu'] > 1
                      and self.resources.available['memory'] > 1
                      and (self.resources.available['gpu'] or not self.resources.total['gpu'])):
                    logger.info('resources available, so request a task')
                    break

        # last update for pilot state
        await self.rpc.update_pilot(self.pilot_id, tasks=[],
                                    resources_available=self.resources.get_available(),
                                    resources_claimed=self.resources.get_claimed())

        if task_errors < 1:
            logger.critical('too many errors when running tasks')
            raise RuntimeError('too many errors')
        else:
            logger.warning('cleanly stopping pilot')

    async def create_task(self, config):
        """
        Create a new Task and start running it

        Args:
            config (dict): The task config
        """
        task_id = config['options']['task_id']

        # add grid-specific config
        for k in self.config['options']:
            if k == 'resources':
                pass
            elif k not in config['options']:
                config['options'][k] = self.config['options'][k]

        tmpdir = tempfile.mkdtemp(suffix='.{}'.format(task_id), dir=os.getcwd())
        config['options']['subprocess_dir'] = tmpdir

        # start the task
        # r = config['options']['resources']
        async for proc in self.runner(config, resources=self.resources):
            ps = psutil.Process(proc.pid) if psutil else None
            self.resources.register_process(task_id, ps, tmpdir)
            data = {
                'p': proc,
                'process': ps,
                'tmpdir': tmpdir,
                'config': config,
            }
            yield data

    def clean_task(self, task_id):
        """Clean up a Task.

        Delete remaining processes and the task temp dir. Release resources
        back to the pilot.

        Args:
            task_id (str): the task_id
        """
        logger.info('cleaning task %s', task_id)
        if task_id in self.tasks:
            task = self.tasks[task_id]
            del self.tasks[task_id]

            # kill process if still running
            try:
                if psutil:
                    # kill children correctly
                    try:
                        processes = task['process'].children(recursive=True)
                    except psutil.NoSuchProcess:
                        pass  # process already died
                    else:
                        processes.reverse()
                        processes.append(task['process'])
                        for p in processes:
                            try:
                                p.terminate()
                            except psutil.NoSuchProcess:
                                pass
                            except Exception:
                                logger.warning('error terminating process',
                                               exc_info=True)

                        def on_terminate(proc):
                            logger.info("process %r terminated with exit code %r",
                                        proc, proc.returncode)
                        try:
                            gone, alive = psutil.wait_procs(processes, timeout=0.1,
                                                            callback=on_terminate)
                            for p in alive:
                                try:
                                    p.kill()
                                except psutil.NoSuchProcess:
                                    pass
                                except Exception:
                                    logger.warning('error killing process',
                                                   exc_info=True)
                        except Exception:
                            logger.warning('failed to kill processes',
                                           exc_info=True)
                task['p'].kill()
            except ProcessLookupError:
                pass  # process already died
            except Exception:
                logger.warning('error deleting process', exc_info=True)

            # copy stdout/stderr
            try:
                os.rename(os.path.join(task['tmpdir'],constants['stderr']), constants['stderr'])
                os.rename(os.path.join(task['tmpdir'],constants['stdout']), constants['stdout'])
            except Exception:
                logger.warning('error copying std[out,err] files', exc_info=True)

            # clean tmpdir
            try:
                if not self.debug:
                    shutil.rmtree(task['tmpdir'])
            except Exception:
                logger.warning('error deleting tmpdir', exc_info=True)

        # return resources to pilot
        try:
            self.resources.release(task_id)
        except Exception:
            logger.warning('error releasing resources', exc_info=True)





ROLES = {
    'admin': ['groups=/tokens/IceProdAdmins'],
    'user': ['groups=/institutions/IceCube.*'],
    'system': ['resource_access.iceprod.roles=iceprod-system'],
}

GROUPS = {
    'admin': ['groups=/tokens/IceProdAdmins'],
    'simprod': ['groups=/posix/simprod-submit'],
    'filtering': ['groups=/posix/i3filter'],
    'users': ['groups=/institutions/IceCube.*'],
}

GROUP_PRIORITIES = {
    'admin': 1.,
    'simprod': .5,
    'filtering': .9,
    'users': .7,
}



"""
Monitor the pilots.

Send monitoring data to graphite.
"""

import argparse
import asyncio
import logging
import os
from collections import defaultdict,Counter

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.core.resources import Resources
from iceprod.server.module import FakeStatsClient, StatsClientIgnoreErrors

logger = logging.getLogger('pilot_monitor')


async def run(rest_client, statsd, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        statsd (:py:class:`statsd.StatsClient`): statsd (graphite) client
        debug (bool): debug flag to propagate exceptions
    """
    try:
        pilots = await rest_client.request('GET', '/pilots')
        res = defaultdict(Counter)
        count = 0
        for pilot in pilots.values():
            for n in ('available','claimed'):
                n2 = 'resources_'+n
                if n2 in pilot:
                    for t in Resources.defaults:
                        if t in pilot[n2] and isinstance(pilot[n2][t], (int,float)) and pilot[n2][t] > 0:
                            res[n][t] += pilot[n2][t]
            count += len(pilot['tasks'])
        for n in res:
            for t in res[n]:
                statsd.gauge('pilot_resources.{}.{}'.format(n,t), int(res[n][t]))
                logger.info('pilot_resources.{}.{} = {}'.format(n,t,res[n][t]))
        statsd.gauge('pilot_count', count)
    except Exception:
        logger.error('error monitoring pilots', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--statsd-address', default=os.environ.get('STATSD_ADDRESS', None), help='statsd address')
    parser.add_argument('--statsd-prefix', default=os.environ.get('STATSD_PREFIX', 'site'), help='statsd site prefix')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    if not args.statsd_address:
        statsd = FakeStatsClient()
    else:
        addr = args.statsd_address
        port = 8125
        if ':' in addr:
            addr,port = addr.split(':')
            port = int(port)
        statsd = StatsClientIgnoreErrors(addr, port=port, prefix=args.statsd_prefix+'.schedule')

    asyncio.run(run(rest_client, statsd, debug=args.debug))


if __name__ == '__main__':
    main()
